import datetime
import json
from enum import Enum
from typing import ClassVar
from uuid import UUID

from pydantic import Field, field_serializer, field_validator

from gen_epix import fastapp
from gen_epix.commondb.domain import enum
from gen_epix.commondb.domain.model.base import Model
from gen_epix.commondb.util import copy_model_field
from gen_epix.fastapp.domain import Entity, create_keys, create_links


class Organization(Model):
    """
    Represents an organization.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="organizations",
        table_name="organization",
        persistable=True,
        keys=create_keys({1: "name", 2: "legal_entity_code"}),
    )
    name: str = Field(
        description="The name of the organization, UNIQUE", max_length=255
    )
    legal_entity_code: str = Field(
        description="The legal entity code of the organization, UNIQUE", max_length=255
    )


class UserNameEmail(fastapp.Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="user_name_emails",
        persistable=False,
    )
    id: UUID | None = Field(  # pyright: ignore[reportIncompatibleVariableOverride]
        default=None, description="The ID of the user"
    )
    name: str | None = Field(
        default=None, description="The full name of the user", max_length=255
    )
    email: str | None = Field(description="The email of the user", max_length=320)


class User(fastapp.User, Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="users",
        table_name="user",
        persistable=True,
        keys=create_keys({1: "key"}),
        links=create_links(
            {
                1: ("organization_id", Organization, "organization"),
            }
        ),
    )
    id: UUID | None = Field(
        default=None, description="The ID of the user"
    )  # pyright: ignore[reportIncompatibleVariableOverride]
    key: str = Field(description="The key of the user, lowercase, UNIQUE", max_length=320)
    email: str | None = Field(
        default=None, description="The email of the user", max_length=320
    )
    name: str | None = Field(
        default=None, description="The full name of the user", max_length=255
    )
    is_active: bool = Field(
        default=True,
        description="Whether the user is active or not. An inactive user cannot perform any actions that require authorization.",
    )
    roles: set[str] = Field(description="The roles of the user", min_length=1)
    organization_id: UUID = Field(
        description="The ID of the organization of the user. FOREIGN KEY"
    )
    organization: Organization | None = Field(
        default=None, description="The organization of the user"
    )

    @field_validator("key", mode="before")
    @classmethod
    def _validate_key(cls, value: str) -> str:
        return value.lower()

    def get_key(self) -> str:
        return self.key

    @field_validator("roles", mode="before")
    @classmethod
    def _validate_roles(cls, value: set[str] | list[str] | str) -> set[str]:
        """
        Validate and convert roles representation to a set[str]. When given as a
        string, it is assumed to be a JSON list.
        """
        if isinstance(value, str):
            return set(json.loads(value))
        if isinstance(value, list):
            return set(value)
        return value

    @field_serializer("roles", mode="plain")
    def _serialize_roles(self, value: set[str]) -> list[str]:
        return list(value)


class OrganizationSet(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="organization_sets",
        table_name="organization_set",
        persistable=True,
        keys=create_keys({1: "name"}),
    )
    name: str = Field(description="The name of the organization set", max_length=255)
    description: str | None = Field(
        None, description="The description of the organization set."
    )


class OrganizationSetMember(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="organization_set_members",
        table_name="organization_set_member",
        persistable=True,
        keys=create_keys({1: ("organization_set_id", "organization_id")}),
        links=create_links(
            {
                1: ("organization_set_id", OrganizationSet, "organization_set"),
                2: ("organization_id", Organization, "organization"),
            }
        ),
    )
    organization_set_id: UUID = Field(
        description="The ID of the organization set. FOREIGN KEY"
    )
    organization_set: OrganizationSet | None = Field(
        default=None, description="The organization set"
    )
    organization_id: UUID = Field(description="The ID of the organization. FOREIGN KEY")
    organization: Organization | None = Field(None, description="The organization")


class Site(Model):
    """
    Represents a physical site of an organization.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="sites",
        table_name="site",
        persistable=True,
        keys=create_keys({1: ("organization_id", "name")}),
        links=create_links(
            {
                1: ("organization_id", Organization, "organization"),
            }
        ),
    )
    organization_id: UUID = Field(description="The ID of the organization. FOREIGN KEY")
    organization: Organization | None = Field(
        default=None, description="The organization corresponding to the ID"
    )
    name: str = Field(description="The name of an organization, UNIQUE", max_length=255)


class Contact(Model):
    """
    A class representing contact information for an organization.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="contacts",
        table_name="contact",
        persistable=True,
        keys=create_keys({1: ("site_id", "name")}),
        links=create_links(
            {
                1: ("site_id", Site, "site"),
            }
        ),
    )
    # TODO: Temporary implementation, check established models for this
    site_id: UUID = Field(
        description="The ID of the site in case the contact is site-specific. FOREIGN KEY",
    )
    site: Site | None = Field(
        default=None, description="The site corresponding to the ID"
    )
    name: str = Field(description="The name of the contact, UNIQUE", max_length=255)
    email: str | None = Field(
        default=None, description="The email address of the contact", max_length=320
    )
    phone: str | None = Field(
        default=None, description="The phone number of the contact"
    )


class IdentifierIssuer(Model):
    """
    A system or process that issues identifiers.
    The combination (identifier_issuer, issued_identifier) is universally unique.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="identifier_issuers",
        table_name="identifier_issuer",
        persistable=True,
    )
    name: str = Field(description="The name of the issuer", max_length=255)


class DataCollection(Model):
    """
    Represents a collection of data.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="data_collections",
        table_name="data_collection",
        persistable=True,
        keys=create_keys({1: "name"}),
    )
    # TODO: Placeholder
    name: str = Field(
        description="The name of a data collection, UNIQUE", max_length=255
    )
    description: str | None = Field(
        default=None, description="The description of the data collection."
    )


class DataCollectionSet(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="data_collection_sets",
        table_name="data_collection_set",
        persistable=True,
        keys=create_keys({1: "name"}),
    )
    name: str = Field(description="The name of the data collection set", max_length=255)
    description: str | None = Field(
        default=None, description="The description of the data collection set."
    )


class DataCollectionSetMember(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="data_collection_set_members",
        table_name="data_collection_set_member",
        persistable=True,
        keys=create_keys(
            {1: ("data_collection_set_id", "data_collection_id")},
        ),
        links=create_links(
            {
                1: ("data_collection_set_id", DataCollectionSet, "data_collection_set"),
                2: ("data_collection_id", DataCollection, "data_collection"),
            }
        ),
    )
    data_collection_set_id: UUID = Field(
        description="The ID of the data collection set. FOREIGN KEY"
    )
    data_collection_set: DataCollectionSet | None = Field(
        default=None, description="The data collection set"
    )
    data_collection_id: UUID = Field(
        description="The ID of the data collection. FOREIGN KEY"
    )
    data_collection: DataCollection | None = Field(
        default=None, description="The data collection"
    )


class UserInvitation(Model):
    """
    Represents an invitation for a new user of a particular organization and
    with particular starting properties.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="user_invitations",
        table_name="user_invitation",
        persistable=True,
        keys=create_keys({1: ("key", "expires_at")}),
        links=create_links(
            {
                1: ("organization_id", Organization, "organization"),
                2: ("invited_by_user_id", User, "invited_by_user"),
            }
        ),
    )
    ROLE_ENUM: ClassVar[type[Enum]] = enum.Role
    key: str = copy_model_field(User, "key")
    email: str | None = copy_model_field(User, "email")
    name: str | None = copy_model_field(User, "name")
    token: str = Field(description="The token of the invitation", max_length=255)
    expires_at: datetime.datetime = Field(
        description="The expiry date of the invitation"
    )
    roles: set[str] = Field(
        description="The initial roles that the new user will have", min_length=1
    )
    invited_by_user_id: UUID = Field(
        description="The ID of the user who invited the new user. FOREIGN KEY"
    )
    invited_by_user: User | None = Field(
        default=None, description="The user who invited the new user"
    )
    organization_id: UUID = Field(
        description="The ID of the organization that the new user will belong to. FOREIGN KEY"
    )
    organization: Organization | None = Field(
        default=None, description="The organization that the new user will belong to"
    )

    @field_validator("roles", mode="before")
    @classmethod
    def _validate_roles(cls, value: set[str] | list[str] | str) -> set[str]:
        """
        Validate and convert roles representation to a set[str]. When given as a
        string, it is assumed to be a JSON list.
        """
        if isinstance(value, str):
            return set(json.loads(value))
        if isinstance(value, list):
            return set(value)
        return value

    @field_serializer("roles", mode="plain")
    def _serialize_roles(self, value: set[str]) -> list[str]:
        return list(value)


class UserInvitationConstraints(Model):
    """
    Represents the constraints for a user invitation.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="user_invitation_constraints",
        persistable=False,
    )
    ROLE_ENUM: ClassVar[type[Enum]] = enum.Role
    roles: set[str] = Field(
        description="The roles that the user may be assigned by the inviting user."
    )
    organization_ids: set[UUID] = Field(
        description="The organizations that the user may be assigned by the inviting user."
    )



# pylint: disable=useless-import-alias
import gen_epix.fastapp as fastapp
from gen_epix.commondb.domain import enum
from gen_epix.commondb.domain.model.abac import (
    OrganizationAdminPolicy as OrganizationAdminPolicy,
)
from gen_epix.commondb.domain.model.base import Model as Model
from gen_epix.commondb.domain.model.organization import Contact as Contact
from gen_epix.commondb.domain.model.organization import DataCollection as DataCollection
from gen_epix.commondb.domain.model.organization import (
    DataCollectionSet as DataCollectionSet,
)
from gen_epix.commondb.domain.model.organization import (
    DataCollectionSetMember as DataCollectionSetMember,
)
from gen_epix.commondb.domain.model.organization import (
    IdentifierIssuer as IdentifierIssuer,
)
from gen_epix.commondb.domain.model.organization import Organization as Organization
from gen_epix.commondb.domain.model.organization import (
    OrganizationSet as OrganizationSet,
)
from gen_epix.commondb.domain.model.organization import (
    OrganizationSetMember as OrganizationSetMember,
)
from gen_epix.commondb.domain.model.organization import Site as Site
from gen_epix.commondb.domain.model.organization import User as User
from gen_epix.commondb.domain.model.organization import UserInvitation as UserInvitation
from gen_epix.commondb.domain.model.organization import (
    UserInvitationConstraints as UserInvitationConstraints,
)
from gen_epix.commondb.domain.model.organization import UserNameEmail as UserNameEmail
from gen_epix.commondb.domain.model.system import Outage as Outage
from gen_epix.commondb.domain.model.system import PackageMetadata as PackageMetadata
from gen_epix.fastapp.services.auth import IdentityProvider as IdentityProvider
from gen_epix.fastapp.services.auth import IDPUser as IDPUser

SORTED_MODELS_BY_SERVICE_TYPE: dict[
    enum.ServiceType, tuple[type[fastapp.Model], ...]
] = {
    enum.ServiceType.AUTH: (
        IdentityProvider,
        IDPUser,
    ),
    enum.ServiceType.SYSTEM: (Outage, PackageMetadata),
    enum.ServiceType.ORGANIZATION: (
        Organization,
        OrganizationSet,
        OrganizationSetMember,
        DataCollection,
        DataCollectionSet,
        DataCollectionSetMember,
        IdentifierIssuer,
        Site,
        Contact,
        UserNameEmail,
        User,
        UserInvitation,
        UserInvitationConstraints,
    ),
    enum.ServiceType.RBAC: tuple(),
    enum.ServiceType.ABAC: (OrganizationAdminPolicy,),
}

SORTED_SERVICE_TYPES = tuple(SORTED_MODELS_BY_SERVICE_TYPE.keys())



# pylint: disable=useless-import-alias
from gen_epix import fastapp
from gen_epix.casedb.domain import enum
from gen_epix.casedb.domain.command.abac import (
    OrganizationAccessCasePolicyCrudCommand as OrganizationAccessCasePolicyCrudCommand,
)
from gen_epix.casedb.domain.command.abac import (
    OrganizationShareCasePolicyCrudCommand as OrganizationShareCasePolicyCrudCommand,
)
from gen_epix.casedb.domain.command.abac import (
    UserAccessCasePolicyCrudCommand as UserAccessCasePolicyCrudCommand,
)
from gen_epix.casedb.domain.command.abac import (
    UserShareCasePolicyCrudCommand as UserShareCasePolicyCrudCommand,
)
from gen_epix.casedb.domain.command.case import CaseCrudCommand as CaseCrudCommand
from gen_epix.casedb.domain.command.case import (
    CaseDataCollectionLinkCrudCommand as CaseDataCollectionLinkCrudCommand,
)
from gen_epix.casedb.domain.command.case import (
    CaseSetCategoryCrudCommand as CaseSetCategoryCrudCommand,
)
from gen_epix.casedb.domain.command.case import CaseSetCrudCommand as CaseSetCrudCommand
from gen_epix.casedb.domain.command.case import (
    CaseSetDataCollectionLinkCrudCommand as CaseSetDataCollectionLinkCrudCommand,
)
from gen_epix.casedb.domain.command.case import (
    CaseSetMemberCrudCommand as CaseSetMemberCrudCommand,
)
from gen_epix.casedb.domain.command.case import (
    CaseSetStatusCrudCommand as CaseSetStatusCrudCommand,
)
from gen_epix.casedb.domain.command.case import (
    CaseTypeColCrudCommand as CaseTypeColCrudCommand,
)
from gen_epix.casedb.domain.command.case import (
    CaseTypeColSetCaseTypeColUpdateAssociationCommand as CaseTypeColSetCaseTypeColUpdateAssociationCommand,
)
from gen_epix.casedb.domain.command.case import (
    CaseTypeColSetCrudCommand as CaseTypeColSetCrudCommand,
)
from gen_epix.casedb.domain.command.case import (
    CaseTypeColSetMemberCrudCommand as CaseTypeColSetMemberCrudCommand,
)
from gen_epix.casedb.domain.command.case import (
    CaseTypeCrudCommand as CaseTypeCrudCommand,
)
from gen_epix.casedb.domain.command.case import (
    CaseTypeSetCaseTypeUpdateAssociationCommand as CaseTypeSetCaseTypeUpdateAssociationCommand,
)
from gen_epix.casedb.domain.command.case import (
    CaseTypeSetCategoryCrudCommand as CaseTypeSetCategoryCrudCommand,
)
from gen_epix.casedb.domain.command.case import (
    CaseTypeSetCrudCommand as CaseTypeSetCrudCommand,
)
from gen_epix.casedb.domain.command.case import (
    CaseTypeSetMemberCrudCommand as CaseTypeSetMemberCrudCommand,
)
from gen_epix.casedb.domain.command.case import ColCrudCommand as ColCrudCommand
from gen_epix.casedb.domain.command.case import CreateCasesCommand as CreateCasesCommand
from gen_epix.casedb.domain.command.case import (
    CreateCaseSetCommand as CreateCaseSetCommand,
)
from gen_epix.casedb.domain.command.case import (
    CreateFileForReadSetCommand as CreateFileForReadSetCommand,
)
from gen_epix.casedb.domain.command.case import (
    CreateFileForSeqCommand as CreateFileForSeqCommand,
)
from gen_epix.casedb.domain.command.case import (
    CreateReadSetsForCasesCommand as CreateReadSetsForCasesCommand,
)
from gen_epix.casedb.domain.command.case import (
    CreateSeqsForCasesCommand as CreateSeqsForCasesCommand,
)
from gen_epix.casedb.domain.command.case import DimCrudCommand as DimCrudCommand
from gen_epix.casedb.domain.command.case import (
    GeneticDistanceProtocolCrudCommand as GeneticDistanceProtocolCrudCommand,
)
from gen_epix.casedb.domain.command.case import (
    RetrieveAlleleProfileCommand as RetrieveAlleleProfileCommand,
)
from gen_epix.casedb.domain.command.case import (
    RetrieveAssemblyProtocolsCommand as RetrieveAssemblyProtocolsCommand,
)
from gen_epix.casedb.domain.command.case import (
    RetrieveCaseRightsCommand as RetrieveCaseRightsCommand,
)
from gen_epix.casedb.domain.command.case import (
    RetrieveCasesByIdCommand as RetrieveCasesByIdCommand,
)
from gen_epix.casedb.domain.command.case import (
    RetrieveCasesByQueryCommand as RetrieveCasesByQueryCommand,
)
from gen_epix.casedb.domain.command.case import (
    RetrieveCaseSetRightsCommand as RetrieveCaseSetRightsCommand,
)
from gen_epix.casedb.domain.command.case import (
    RetrieveCaseSetStatsCommand as RetrieveCaseSetStatsCommand,
)
from gen_epix.casedb.domain.command.case import (
    RetrieveCaseTypeStatsCommand as RetrieveCaseTypeStatsCommand,
)
from gen_epix.casedb.domain.command.case import (
    RetrieveCompleteCaseTypeCommand as RetrieveCompleteCaseTypeCommand,
)
from gen_epix.casedb.domain.command.case import (
    RetrieveGeneticSequenceByCaseCommand as RetrieveGeneticSequenceByCaseCommand,
)
from gen_epix.casedb.domain.command.case import (
    RetrieveGeneticSequenceFastaByCaseCommand as RetrieveGeneticSequenceFastaByCaseCommand,
)
from gen_epix.casedb.domain.command.case import (
    RetrieveLibraryPrepProtocolsCommand as RetrieveLibraryPrepProtocolsCommand,
)
from gen_epix.casedb.domain.command.case import (
    RetrievePhylogeneticTreeByCasesCommand as RetrievePhylogeneticTreeByCasesCommand,
)
from gen_epix.casedb.domain.command.case import (
    RetrievePhylogeneticTreeBySequencesCommand as RetrievePhylogeneticTreeBySequencesCommand,
)
from gen_epix.casedb.domain.command.case import (
    TreeAlgorithmClassCrudCommand as TreeAlgorithmClassCrudCommand,
)
from gen_epix.casedb.domain.command.case import (
    TreeAlgorithmCrudCommand as TreeAlgorithmCrudCommand,
)
from gen_epix.casedb.domain.command.case import (
    ValidateCasesCommand as ValidateCasesCommand,
)
from gen_epix.casedb.domain.command.geo import RegionCrudCommand as RegionCrudCommand
from gen_epix.casedb.domain.command.geo import (
    RegionRelationCrudCommand as RegionRelationCrudCommand,
)
from gen_epix.casedb.domain.command.geo import (
    RegionSetCrudCommand as RegionSetCrudCommand,
)
from gen_epix.casedb.domain.command.geo import (
    RegionSetShapeCrudCommand as RegionSetShapeCrudCommand,
)
from gen_epix.casedb.domain.command.geo import (
    RetrieveContainingRegionCommand as RetrieveContainingRegionCommand,
)
from gen_epix.casedb.domain.command.ontology import (
    ConceptCrudCommand as ConceptCrudCommand,
)
from gen_epix.casedb.domain.command.ontology import (
    ConceptRelationCrudCommand as ConceptRelationCrudCommand,
)
from gen_epix.casedb.domain.command.ontology import (
    ConceptSetCrudCommand as ConceptSetCrudCommand,
)
from gen_epix.casedb.domain.command.ontology import (
    DiseaseCrudCommand as DiseaseCrudCommand,
)
from gen_epix.casedb.domain.command.ontology import (
    DiseaseEtiologicalAgentUpdateAssociationCommand as DiseaseEtiologicalAgentUpdateAssociationCommand,
)
from gen_epix.casedb.domain.command.ontology import (
    EtiologicalAgentCrudCommand as EtiologicalAgentCrudCommand,
)
from gen_epix.casedb.domain.command.ontology import (
    EtiologyCrudCommand as EtiologyCrudCommand,
)
from gen_epix.casedb.domain.command.seqdb import (
    RetrieveGeneticSequenceByIdCommand as RetrieveGeneticSequenceByIdCommand,
)
from gen_epix.casedb.domain.command.seqdb import (
    RetrieveGeneticSequenceFastaByIdCommand as RetrieveGeneticSequenceFastaByIdCommand,
)
from gen_epix.casedb.domain.command.subject import (
    SubjectCrudCommand as SubjectCrudCommand,
)
from gen_epix.casedb.domain.command.subject import (
    SubjectIdentifierCrudCommand as SubjectIdentifierCrudCommand,
)
from gen_epix.commondb.domain import enum as common_enum
from gen_epix.commondb.domain.command import (
    COMMANDS_BY_SERVICE_TYPE as _COMMON_COMMANDS_BY_SERVICE_TYPE,
)
from gen_epix.commondb.domain.command import Command as Command
from gen_epix.commondb.domain.command import ContactCrudCommand as ContactCrudCommand
from gen_epix.commondb.domain.command import CrudCommand as CrudCommand
from gen_epix.commondb.domain.command import (
    DataCollectionCrudCommand as DataCollectionCrudCommand,
)
from gen_epix.commondb.domain.command import (
    DataCollectionSetCrudCommand as DataCollectionSetCrudCommand,
)
from gen_epix.commondb.domain.command import (
    DataCollectionSetDataCollectionUpdateAssociationCommand as DataCollectionSetDataCollectionUpdateAssociationCommand,
)
from gen_epix.commondb.domain.command import (
    DataCollectionSetMemberCrudCommand as DataCollectionSetMemberCrudCommand,
)
from gen_epix.commondb.domain.command import (
    GetIdentityProvidersCommand as GetIdentityProvidersCommand,
)
from gen_epix.commondb.domain.command import (
    IdentifierIssuerCrudCommand as IdentifierIssuerCrudCommand,
)
from gen_epix.commondb.domain.command import InviteUserCommand as InviteUserCommand
from gen_epix.commondb.domain.command import (
    OrganizationAdminPolicyCrudCommand as OrganizationAdminPolicyCrudCommand,
)
from gen_epix.commondb.domain.command import (
    OrganizationCrudCommand as OrganizationCrudCommand,
)
from gen_epix.commondb.domain.command import (
    OrganizationSetCrudCommand as OrganizationSetCrudCommand,
)
from gen_epix.commondb.domain.command import (
    OrganizationSetMemberCrudCommand as OrganizationSetMemberCrudCommand,
)
from gen_epix.commondb.domain.command import (
    OrganizationSetOrganizationUpdateAssociationCommand as OrganizationSetOrganizationUpdateAssociationCommand,
)
from gen_epix.commondb.domain.command import OutageCrudCommand as OutageCrudCommand
from gen_epix.commondb.domain.command import (
    RegisterInvitedUserCommand as RegisterInvitedUserCommand,
)
from gen_epix.commondb.domain.command import (
    RetrieveOrganizationAdminNameEmailsCommand as RetrieveOrganizationAdminNameEmailsCommand,
)
from gen_epix.commondb.domain.command import (
    RetrieveOrganizationContactCommand as RetrieveOrganizationContactCommand,
)
from gen_epix.commondb.domain.command import (
    RetrieveOrganizationsUnderAdminCommand as RetrieveOrganizationsUnderAdminCommand,
)
from gen_epix.commondb.domain.command import (
    RetrieveOutagesCommand as RetrieveOutagesCommand,
)
from gen_epix.commondb.domain.command import (
    RetrieveOwnPermissionsCommand as RetrieveOwnPermissionsCommand,
)
from gen_epix.commondb.domain.command import SiteCrudCommand as SiteCrudCommand
from gen_epix.commondb.domain.command import (
    UpdateAssociationCommand as UpdateAssociationCommand,
)
from gen_epix.commondb.domain.command import UpdateUserCommand as UpdateUserCommand
from gen_epix.commondb.domain.command import (
    UpdateUserOwnOrganizationCommand as UpdateUserOwnOrganizationCommand,
)
from gen_epix.commondb.domain.command import UserCrudCommand as UserCrudCommand
from gen_epix.commondb.domain.command import (
    UserInvitationCrudCommand as UserInvitationCrudCommand,
)
from gen_epix.commondb.domain.command.organization import (
    RetrieveInviteUserConstraintsCommand as RetrieveInviteUserConstraintsCommand,
)
from gen_epix.commondb.domain.command.rbac import (
    RetrieveSubRolesCommand as RetrieveSubRolesCommand,
)

COMMANDS_BY_SERVICE_TYPE: dict[enum.ServiceType, set[type[fastapp.Command]]] = {
    # Specific commands
    enum.ServiceType.ABAC: set(
        _COMMON_COMMANDS_BY_SERVICE_TYPE[common_enum.ServiceType.ABAC]
    ).union(
        {
            OrganizationAccessCasePolicyCrudCommand,
            OrganizationShareCasePolicyCrudCommand,
            UserAccessCasePolicyCrudCommand,
            UserShareCasePolicyCrudCommand,
        }
    ),
    enum.ServiceType.CASE: {
        CaseCrudCommand,
        CaseDataCollectionLinkCrudCommand,
        CreateCasesCommand,
        CaseSetCategoryCrudCommand,
        CreateCaseSetCommand,
        CaseSetCrudCommand,
        CaseSetDataCollectionLinkCrudCommand,
        CaseSetMemberCrudCommand,
        CaseSetStatusCrudCommand,
        CaseTypeColCrudCommand,
        CaseTypeColSetCaseTypeColUpdateAssociationCommand,
        CaseTypeColSetCrudCommand,
        CaseTypeColSetMemberCrudCommand,
        CaseTypeCrudCommand,
        CaseTypeSetCaseTypeUpdateAssociationCommand,
        CaseTypeSetCategoryCrudCommand,
        CaseTypeSetCrudCommand,
        CaseTypeSetMemberCrudCommand,
        ColCrudCommand,
        DimCrudCommand,
        GeneticDistanceProtocolCrudCommand,
        RetrieveAlleleProfileCommand,
        CreateReadSetsForCasesCommand,
        CreateSeqsForCasesCommand,
        CreateFileForReadSetCommand,
        CreateFileForSeqCommand,
        RetrieveCaseRightsCommand,
        RetrieveCasesByIdCommand,
        RetrieveCasesByQueryCommand,
        RetrieveCaseSetRightsCommand,
        RetrieveCaseSetStatsCommand,
        RetrieveCaseTypeStatsCommand,
        RetrieveCompleteCaseTypeCommand,
        RetrieveGeneticSequenceByCaseCommand,
        RetrievePhylogeneticTreeByCasesCommand,
        RetrievePhylogeneticTreeBySequencesCommand,
        RetrieveGeneticSequenceFastaByCaseCommand,
        TreeAlgorithmClassCrudCommand,
        TreeAlgorithmCrudCommand,
        ValidateCasesCommand,
        RetrieveLibraryPrepProtocolsCommand,
        RetrieveAssemblyProtocolsCommand,
    },
    enum.ServiceType.GEO: {
        RegionCrudCommand,
        RegionRelationCrudCommand,
        RegionSetCrudCommand,
        RegionSetShapeCrudCommand,
        RetrieveContainingRegionCommand,
    },
    enum.ServiceType.ONTOLOGY: {
        ConceptSetCrudCommand,
        ConceptCrudCommand,
        ConceptRelationCrudCommand,
        DiseaseCrudCommand,
        DiseaseEtiologicalAgentUpdateAssociationCommand,
        EtiologicalAgentCrudCommand,
        EtiologyCrudCommand,
    },
    enum.ServiceType.SEQDB: {
        RetrieveGeneticSequenceByIdCommand,
        RetrieveGeneticSequenceFastaByIdCommand,
    },
    enum.ServiceType.SUBJECT: {
        SubjectCrudCommand,
        SubjectIdentifierCrudCommand,
    },
    # Common commands
    enum.ServiceType.AUTH: set(
        _COMMON_COMMANDS_BY_SERVICE_TYPE[common_enum.ServiceType.AUTH]
    ),
    enum.ServiceType.SYSTEM: set(
        _COMMON_COMMANDS_BY_SERVICE_TYPE[common_enum.ServiceType.SYSTEM]
    ),
    enum.ServiceType.RBAC: set(
        _COMMON_COMMANDS_BY_SERVICE_TYPE[common_enum.ServiceType.RBAC]
    ),
    enum.ServiceType.ORGANIZATION: set(
        _COMMON_COMMANDS_BY_SERVICE_TYPE[common_enum.ServiceType.ORGANIZATION]
    ),
}

COMMON_COMMAND_MAP: dict[type[fastapp.Command], type[fastapp.Command]] = {}



import abc
import uuid

from gen_epix.commondb.domain import command, model
from gen_epix.commondb.domain.enum import ServiceType
from gen_epix.commondb.domain.repository import BaseAbacRepository
from gen_epix.fastapp import BaseService
from gen_epix.fastapp.model import Command


class BaseAbacService(BaseService):
    SERVICE_TYPE = ServiceType.ABAC

    ORGANIZATION_ADMIN_WRITE_COMMANDS: set[type[Command]] = {
        command.ContactCrudCommand,
        command.SiteCrudCommand,
    }

    READ_USER_COMMANDS: set[type[Command]] = {
        command.UserCrudCommand,
    }

    UPDATE_USER_COMMANDS: set[type[Command]] = {
        command.InviteUserCommand,
        command.UpdateUserCommand,
    }

    READ_ORGANIZATION_RESULTS_ONLY_COMMANDS: set[type[Command]] = {
        command.OrganizationAdminPolicyCrudCommand,
        command.UserInvitationCrudCommand,
        command.RetrieveInviteUserConstraintsCommand,
    }

    READ_SELF_RESULTS_ONLY_COMMANDS: set[type[Command]] = set()

    # Property overridden to provide narrower return value to support linter
    @property  # type: ignore
    def repository(self) -> BaseAbacRepository:  # type: ignore
        return super().repository  # type: ignore

    def register_handlers(self) -> None:
        self.register_default_crud_handlers()
        f = self.app.register_handler
        f(
            command.RetrieveOrganizationAdminNameEmailsCommand,
            self.retrieve_organization_admin_name_emails,
        )
        f(
            command.RetrieveOrganizationsUnderAdminCommand,
            self.retrieve_organizations_under_admin,
        )
        f(
            command.UpdateUserOwnOrganizationCommand,
            self.temp_update_user_own_organization,
        )

    @abc.abstractmethod
    def retrieve_organization_admin_name_emails(
        self,
        cmd: command.RetrieveOrganizationAdminNameEmailsCommand,
    ) -> list[model.UserNameEmail]:
        raise NotImplementedError

    @abc.abstractmethod
    def retrieve_organizations_under_admin(
        self, cmd: command.RetrieveOrganizationsUnderAdminCommand
    ) -> set[uuid.UUID]:
        raise NotImplementedError

    def temp_update_user_own_organization(
        self,
        cmd: command.UpdateUserOwnOrganizationCommand,
    ) -> model.User:
        raise NotImplementedError



from uuid import UUID

import sqlalchemy.orm as orm
from sqlalchemy.orm import Mapped

from gen_epix.casedb.domain import DOMAIN, enum, model
from gen_epix.commondb.repositories.sa_model import (
    RowMetadataMixin,
    create_mapped_column,
    create_table_args,
)

Base: type = orm.declarative_base(name=enum.ServiceType.ABAC.value)


class OrganizationAccessCasePolicy(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(
        model.OrganizationAccessCasePolicy
    )

    organization_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.OrganizationAccessCasePolicy, "organization_id"
    )
    data_collection_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.OrganizationAccessCasePolicy, "data_collection_id"
    )
    case_type_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.OrganizationAccessCasePolicy, "case_type_set_id"
    )
    is_active: Mapped[bool] = create_mapped_column(
        DOMAIN, model.OrganizationAccessCasePolicy, "is_active"
    )
    is_private: Mapped[bool] = create_mapped_column(
        DOMAIN, model.OrganizationAccessCasePolicy, "is_private"
    )
    add_case: Mapped[bool] = create_mapped_column(
        DOMAIN, model.OrganizationAccessCasePolicy, "add_case"
    )
    remove_case: Mapped[bool] = create_mapped_column(
        DOMAIN, model.OrganizationAccessCasePolicy, "remove_case"
    )
    add_case_set: Mapped[bool] = create_mapped_column(
        DOMAIN, model.OrganizationAccessCasePolicy, "add_case_set"
    )
    remove_case_set: Mapped[bool] = create_mapped_column(
        DOMAIN, model.OrganizationAccessCasePolicy, "remove_case_set"
    )
    read_case_type_col_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.OrganizationAccessCasePolicy, "read_case_type_col_set_id"
    )
    write_case_type_col_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.OrganizationAccessCasePolicy, "write_case_type_col_set_id"
    )
    read_case_set: Mapped[bool] = create_mapped_column(
        DOMAIN, model.OrganizationAccessCasePolicy, "read_case_set"
    )
    write_case_set: Mapped[bool] = create_mapped_column(
        DOMAIN, model.OrganizationAccessCasePolicy, "write_case_set"
    )


class UserAccessCasePolicy(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.UserAccessCasePolicy)

    user_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "user_id"
    )
    data_collection_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "data_collection_id"
    )
    case_type_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "case_type_set_id"
    )
    is_active: Mapped[bool] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "is_active"
    )
    add_case: Mapped[bool] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "add_case"
    )
    remove_case: Mapped[bool] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "remove_case"
    )
    add_case_set: Mapped[bool] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "add_case_set"
    )
    remove_case_set: Mapped[bool] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "remove_case_set"
    )
    read_case_type_col_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "read_case_type_col_set_id"
    )
    write_case_type_col_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "write_case_type_col_set_id"
    )
    read_case_set: Mapped[bool] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "read_case_set"
    )
    write_case_set: Mapped[bool] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "write_case_set"
    )


class OrganizationShareCasePolicy(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.OrganizationShareCasePolicy)

    organization_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.OrganizationShareCasePolicy, "organization_id"
    )
    data_collection_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.OrganizationShareCasePolicy, "data_collection_id"
    )
    case_type_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.OrganizationShareCasePolicy, "case_type_set_id"
    )
    from_data_collection_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.OrganizationShareCasePolicy, "from_data_collection_id"
    )
    is_active: Mapped[bool] = create_mapped_column(
        DOMAIN, model.OrganizationShareCasePolicy, "is_active"
    )
    add_case: Mapped[bool] = create_mapped_column(
        DOMAIN, model.OrganizationShareCasePolicy, "add_case"
    )
    remove_case: Mapped[bool] = create_mapped_column(
        DOMAIN, model.OrganizationShareCasePolicy, "remove_case"
    )
    add_case_set: Mapped[bool] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "add_case_set"
    )
    remove_case_set: Mapped[bool] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "remove_case_set"
    )


class UserShareCasePolicy(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.UserShareCasePolicy)

    user_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.UserShareCasePolicy, "user_id"
    )
    data_collection_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.UserShareCasePolicy, "data_collection_id"
    )
    case_type_set_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.UserShareCasePolicy, "case_type_set_id"
    )
    from_data_collection_id: Mapped[UUID] = create_mapped_column(
        DOMAIN, model.UserShareCasePolicy, "from_data_collection_id"
    )
    is_active: Mapped[bool] = create_mapped_column(
        DOMAIN, model.UserShareCasePolicy, "is_active"
    )
    add_case: Mapped[bool] = create_mapped_column(
        DOMAIN, model.UserShareCasePolicy, "add_case"
    )
    remove_case: Mapped[bool] = create_mapped_column(
        DOMAIN, model.UserShareCasePolicy, "remove_case"
    )
    add_case_set: Mapped[bool] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "add_case_set"
    )
    remove_case_set: Mapped[bool] = create_mapped_column(
        DOMAIN, model.UserAccessCasePolicy, "remove_case_set"
    )



"""
Validation transformer implementation.
"""

from collections.abc import Callable

from gen_epix.transform.adapter import ObjectAdapter
from gen_epix.transform.transformer import Transformer


class ValidationTransformer(Transformer):
    """Validate object and fail if validation doesn't pass."""

    def __init__(
        self, validator: Callable[[ObjectAdapter], bool], name: str | None = None
    ):
        super().__init__(name)
        self.validator = validator

    def transform(self, obj: ObjectAdapter) -> ObjectAdapter:
        """Validate object and raise exception if validation fails."""
        if not self.validator(obj):
            raise ValueError(f"Validation failed for object: {obj.unwrap()}")
        return obj



from abc import abstractmethod
from collections.abc import Iterable
from typing import Any
from uuid import UUID

import gen_epix.casedb.domain.command as command
import gen_epix.casedb.domain.enum as enum
import gen_epix.casedb.domain.model as model
from gen_epix.casedb.domain.service import BaseCaseService as DomainBaseCaseService
from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.fastapp import BaseUnitOfWork
from gen_epix.filter import DatetimeRangeFilter, Filter


class BaseCaseService(DomainBaseCaseService):
    """
    Abstract base class for case services defining the interface contract.
    This additional base class allows splitting the implementation into
    multiple modules while maintaining linter support.
    """

    _VALUE_TO_STR = {
        enum.ColType.TIME_DAY: lambda x: None if not x else f"{x}",
        enum.ColType.TIME_WEEK: lambda x: None if not x else f"{x}",
        enum.ColType.TIME_MONTH: lambda x: None if not x else f"{x}",
        enum.ColType.TIME_QUARTER: lambda x: None if not x else f"{x}",
        enum.ColType.TIME_YEAR: lambda x: None if not x else f"{x}",
        enum.ColType.GEO_LATLON: lambda x: None if not x else f"{x}",
        enum.ColType.TEXT: lambda x: None if not x else f"{x}",
        enum.ColType.ID_DIRECT: lambda x: None if not x else f"{x}",
        enum.ColType.ID_PSEUDONYMISED: lambda x: None if not x else f"{x}",
        enum.ColType.OTHER: lambda x: None if not x else f"{x}",
        enum.ColType.DECIMAL_0: lambda x: (
            None if not x else (x if isinstance(x, str) else f"{x:.0f}")
        ),
        enum.ColType.DECIMAL_1: lambda x: (
            None if not x else (x if isinstance(x, str) else f"{x:.1f}")
        ),
        enum.ColType.DECIMAL_2: lambda x: (
            None if not x else (x if isinstance(x, str) else f"{x:.2f}")
        ),
        enum.ColType.DECIMAL_3: lambda x: (
            None if not x else (x if isinstance(x, str) else f"{x:.3f}")
        ),
        enum.ColType.DECIMAL_4: lambda x: (
            None if not x else (x if isinstance(x, str) else f"{x:.4f}")
        ),
        enum.ColType.DECIMAL_5: lambda x: (
            None if not x else (x if isinstance(x, str) else f"{x:.5f}")
        ),
        enum.ColType.DECIMAL_6: lambda x: (
            None if not x else (x if isinstance(x, str) else f"{x:.6f}")
        ),
    }

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        app_impl: AppImplDetails = self.app.impl
        self.role_map = app_impl.role_map
        self.role_set_map = app_impl.role_set_map

    @abstractmethod
    def crud(  # type:ignore[override]
        self, cmd: command.CrudCommand
    ) -> list[model.Model] | model.Model | list[UUID] | UUID | list[bool] | bool | None:
        """
        Override the base crud method to apply ABAC restrictions and cascade delete
        where necessary
        """
        raise NotImplementedError()

    @abstractmethod
    def validate_cases(
        self, cmd: command.ValidateCasesCommand
    ) -> model.CaseValidationReport:
        """Validate cases according to case type rules."""
        raise NotImplementedError()

    @abstractmethod
    def create_cases(self, cmd: command.CreateCasesCommand) -> list[model.Case] | None:
        """Create new cases with validation and authorization checks."""
        raise NotImplementedError()

    @abstractmethod
    def create_case_set(
        self, cmd: command.CreateCaseSetCommand
    ) -> model.CaseSet | None:
        """Create a new case set with associated data collection links."""
        raise NotImplementedError()

    @abstractmethod
    def retrieve_complete_case_type(
        self,
        cmd: command.RetrieveCompleteCaseTypeCommand,
    ) -> model.CompleteCaseType:
        """Retrieve complete case type information including all related metadata."""
        raise NotImplementedError()

    @abstractmethod
    def retrieve_case_type_stats(
        self,
        cmd: command.RetrieveCaseTypeStatsCommand,
    ) -> list[model.CaseTypeStat]:
        """Retrieve statistical information about case types."""
        raise NotImplementedError()

    @abstractmethod
    def retrieve_case_set_stats(
        self,
        cmd: command.RetrieveCaseSetStatsCommand,
    ) -> list[model.CaseSetStat]:
        """Retrieve statistical information about case sets."""
        raise NotImplementedError()

    @abstractmethod
    def retrieve_cases_by_query(
        self, cmd: command.RetrieveCasesByQueryCommand
    ) -> list[UUID]:
        """Retrieve case IDs based on query criteria with ABAC filtering."""
        raise NotImplementedError()

    @abstractmethod
    def retrieve_cases_by_id(
        self, cmd: command.RetrieveCasesByIdCommand
    ) -> list[model.Case]:
        """Retrieve cases by their IDs with content filtering."""
        raise NotImplementedError()

    @abstractmethod
    def retrieve_case_or_set_rights(
        self,
        cmd: command.RetrieveCaseRightsCommand | command.RetrieveCaseSetRightsCommand,
    ) -> list[model.CaseRights] | list[model.CaseSetRights]:
        """Retrieve access rights for cases or case sets."""
        raise NotImplementedError()

    @abstractmethod
    def retrieve_phylogenetic_tree(
        self, cmd: command.RetrievePhylogeneticTreeByCasesCommand
    ) -> model.PhylogeneticTree:
        """Retrieve phylogenetic tree for a set of cases."""
        raise NotImplementedError()

    @abstractmethod
    def retrieve_genetic_sequence_by_case(
        self,
        cmd: command.RetrieveGeneticSequenceByCaseCommand,
    ) -> list[model.GeneticSequence]:
        """Retrieve genetic sequences associated with cases."""
        raise NotImplementedError()

    @abstractmethod
    def retrieve_genetic_sequence_fasta_by_case(
        self, cmd: command.RetrieveGeneticSequenceFastaByCaseCommand
    ) -> Iterable[str]:
        """Return a streaming iterable of FASTA formatted lines for genetic sequences."""
        raise NotImplementedError()

    @abstractmethod
    def _read_association_with_valid_ids(
        self,
        command_class: type[command.CrudCommand],
        field_name1: str,
        field_name2: str,
        valid_ids1: set[UUID] | frozenset[UUID] | None = None,
        valid_ids2: set[UUID] | frozenset[UUID] | None = None,
        match_all1: bool = False,
        match_all2: bool = False,
        return_type: str = "objects",
        uow: BaseUnitOfWork | None = None,
        user: model.User | None = None,
    ) -> list[model.Model] | list[UUID] | dict[UUID, set[UUID]]:
        """Read association entities with ID validation."""
        raise NotImplementedError()

    @abstractmethod
    def _retrieve_case_sets_with_content_right(
        self,
        uow: BaseUnitOfWork,
        user_id: UUID,
        case_abac: model.CaseAbac,
        right: enum.CaseRight,
        case_set_ids: list[UUID] | None = None,
        case_type_ids: set[UUID] | None = None,
        filter: Filter | None = None,
        on_invalid_case_set_id: str = "raise",
    ) -> list[model.CaseSet]:
        """Retrieve case sets that the user has specific content rights for."""
        raise NotImplementedError()

    @abstractmethod
    def _retrieve_cases_with_content_right(
        self,
        uow: BaseUnitOfWork,
        user_id: UUID,
        case_abac: model.CaseAbac,
        right: enum.CaseRight,
        case_ids: list[UUID] | None = None,
        case_type_ids: set[UUID] | None = None,
        datetime_range_filter: DatetimeRangeFilter | None = None,
        on_invalid_case_id: str = "raise",
        filter_content: bool = True,
        extra_access_case_type_col_ids: set[UUID] | None = None,
    ) -> list[model.Case]:
        """Retrieve cases that the user has specific content rights for."""
        raise NotImplementedError()

    @abstractmethod
    def _retrieve_case_data_collections_map(
        self,
        uow: BaseUnitOfWork,
        user_id: UUID,
        case_ids: Iterable[UUID] | None = None,
        data_collection_ids: Iterable[UUID] | None = None,
    ) -> dict[UUID, set[UUID]]:
        """Retrieve mapping of cases to their data collections."""
        raise NotImplementedError()

    @abstractmethod
    def _retrieve_case_set_data_collections_map(
        self,
        uow: BaseUnitOfWork,
        user_id: UUID,
        case_set_ids: Iterable[UUID] | None = None,
        data_collection_ids: Iterable[UUID] | None = None,
    ) -> dict[UUID, set[UUID]]:
        """Retrieve mapping of case sets to their data collections."""
        raise NotImplementedError()

    @abstractmethod
    def _retrieve_case_case_sets_map(
        self,
        uow: BaseUnitOfWork,
        user_id: UUID,
        case_ids: Iterable[UUID] | None = None,
        case_set_ids: Iterable[UUID] | None = None,
    ) -> dict[UUID, set[UUID]]:
        """Retrieve mapping of cases to their case sets."""
        raise NotImplementedError()

    @abstractmethod
    def _retrieve_association_map(
        self,
        uow: BaseUnitOfWork,
        user_id: UUID | None,
        association_class: type[model.Model],
        link_field_name1: str,
        link_field_name2: str,
        obj_ids1: frozenset[UUID] | None = None,
        obj_ids2: frozenset[UUID] | None = None,
    ) -> dict[UUID, set[UUID]]:
        """
        Get a dict[obj_id1, set[obj_ids]] based on the association stored in the association_class objs.
        """
        raise NotImplementedError()

    @abstractmethod
    def _retrieve_sequence_column_data(
        self, uow: BaseUnitOfWork, user: model.User, seq_case_type_col_id: UUID
    ) -> tuple[model.CaseTypeCol, model.Col]:
        """Retrieve sequence column data and validate it's a genetic sequence column."""
        raise NotImplementedError()

    @abstractmethod
    def _verify_case_set_member_case_type(
        self, user: model.User, case_set_members: list[model.CaseSetMember]
    ) -> None:
        """Verify that case set members have matching case types with their case sets."""
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def _compose_id_filter(*key_and_ids: tuple[str, set[UUID]]) -> Filter:
        """Compose filter for ID-based filtering."""
        pass



from typing import ClassVar

import gen_epix.casedb.domain.model.subject as model
from gen_epix.commondb.domain.command import CrudCommand

# Non-CRUD


# CRUD


class SubjectCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Subject


class SubjectIdentifierCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SubjectIdentifier



from enum import Enum


class FilterType(Enum):
    BASE = "BASE"
    EXISTS = "EXISTS"
    EQUALS_BOOLEAN = "EQUALS_BOOLEAN"
    EQUALS_NUMBER = "EQUALS_NUMBER"
    EQUALS_STRING = "EQUALS_STRING"
    EQUALS_UUID = "EQUALS_UUID"
    COMPOSITE = "COMPOSITE"
    DATE_RANGE = "DATE_RANGE"
    DATETIME_RANGE = "DATETIME_RANGE"
    NUMBER_RANGE = "NUMBER_RANGE"
    PARTIAL_DATE_RANGE = "PARTIAL_DATE_RANGE"
    RANGE = "RANGE"
    REGEX = "REGEX"
    NUMBER_SET = "NUMBER_SET"
    STRING_SET = "STRING_SET"
    UUID_SET = "UUID_SET"
    VALUE_SET = "VALUE_SET"
    NO_FILTER = "NO_FILTER"


class LogicalOperator(Enum):
    AND = "AND"
    OR = "OR"
    NOT = "NOT"
    XOR = "XOR"
    NAND = "NAND"
    NOR = "NOR"
    XNOR = "XNOR"
    IMPLIES = "IMPLIES"
    NIMPLIES = "NIMPLIES"


class ComparisonOperator(Enum):
    ST = "<"
    STE = "<="
    EQ = "="
    GTE = ">="
    GT = ">"
    NEQ = "!="



# pylint: disable=useless-import-alias
from gen_epix import fastapp
from gen_epix.commondb.domain import enum as common_enum
from gen_epix.commondb.domain.model import (
    SORTED_MODELS_BY_SERVICE_TYPE as _COMMON_SORTED_MODELS_BY_SERVICE_TYPE,
)
from gen_epix.commondb.domain.model import Contact as Contact
from gen_epix.commondb.domain.model import DataCollection as DataCollection
from gen_epix.commondb.domain.model import DataCollectionSet as DataCollectionSet
from gen_epix.commondb.domain.model import (
    DataCollectionSetMember as DataCollectionSetMember,
)
from gen_epix.commondb.domain.model import IdentifierIssuer as IdentifierIssuer
from gen_epix.commondb.domain.model import Model as Model
from gen_epix.commondb.domain.model import Organization as Organization
from gen_epix.commondb.domain.model import (
    OrganizationAdminPolicy as OrganizationAdminPolicy,
)
from gen_epix.commondb.domain.model import OrganizationSet as OrganizationSet
from gen_epix.commondb.domain.model import (
    OrganizationSetMember as OrganizationSetMember,
)
from gen_epix.commondb.domain.model import Outage as Outage
from gen_epix.commondb.domain.model import Site as Site
from gen_epix.commondb.domain.model import User as User
from gen_epix.commondb.domain.model import UserInvitation as UserInvitation
from gen_epix.commondb.domain.model import (
    UserInvitationConstraints as UserInvitationConstraints,
)
from gen_epix.commondb.domain.model import UserNameEmail as UserNameEmail
from gen_epix.fastapp.services.auth import IdentityProvider as IdentityProvider
from gen_epix.fastapp.services.auth import IDPUser as IDPUser
from gen_epix.omopdb.domain import enum
from gen_epix.omopdb.domain.model.omop import CareSite as CareSite
from gen_epix.omopdb.domain.model.omop import CdmSource as CdmSource
from gen_epix.omopdb.domain.model.omop import Cohort as Cohort
from gen_epix.omopdb.domain.model.omop import CohortDefinition as CohortDefinition
from gen_epix.omopdb.domain.model.omop import Concept as Concept
from gen_epix.omopdb.domain.model.omop import ConceptAncestor as ConceptAncestor
from gen_epix.omopdb.domain.model.omop import ConceptClass as ConceptClass
from gen_epix.omopdb.domain.model.omop import ConceptRelationship as ConceptRelationship
from gen_epix.omopdb.domain.model.omop import ConceptSynonym as ConceptSynonym
from gen_epix.omopdb.domain.model.omop import ConditionEra as ConditionEra
from gen_epix.omopdb.domain.model.omop import ConditionOccurrence as ConditionOccurrence
from gen_epix.omopdb.domain.model.omop import Cost as Cost
from gen_epix.omopdb.domain.model.omop import DataLineageMixin as DataLineageMixin
from gen_epix.omopdb.domain.model.omop import DeviceExposure as DeviceExposure
from gen_epix.omopdb.domain.model.omop import Domain as Domain
from gen_epix.omopdb.domain.model.omop import DoseEra as DoseEra
from gen_epix.omopdb.domain.model.omop import DrugEra as DrugEra
from gen_epix.omopdb.domain.model.omop import DrugExposure as DrugExposure
from gen_epix.omopdb.domain.model.omop import DrugStrength as DrugStrength
from gen_epix.omopdb.domain.model.omop import FactRelationship as FactRelationship
from gen_epix.omopdb.domain.model.omop import Location as Location
from gen_epix.omopdb.domain.model.omop import LocationHistory as LocationHistory
from gen_epix.omopdb.domain.model.omop import Measurement as Measurement
from gen_epix.omopdb.domain.model.omop import MeasurementRelation as MeasurementRelation
from gen_epix.omopdb.domain.model.omop import Metadata as Metadata
from gen_epix.omopdb.domain.model.omop import Note as Note
from gen_epix.omopdb.domain.model.omop import NoteNlp as NoteNlp
from gen_epix.omopdb.domain.model.omop import Observation as Observation
from gen_epix.omopdb.domain.model.omop import ObservationPeriod as ObservationPeriod
from gen_epix.omopdb.domain.model.omop import PayerPlanPeriod as PayerPlanPeriod
from gen_epix.omopdb.domain.model.omop import Person as Person
from gen_epix.omopdb.domain.model.omop import ProcedureOccurrence as ProcedureOccurrence
from gen_epix.omopdb.domain.model.omop import Provider as Provider
from gen_epix.omopdb.domain.model.omop import Relationship as Relationship
from gen_epix.omopdb.domain.model.omop import SourceToConceptMap as SourceToConceptMap
from gen_epix.omopdb.domain.model.omop import Specimen as Specimen
from gen_epix.omopdb.domain.model.omop import SurveyConduct as SurveyConduct
from gen_epix.omopdb.domain.model.omop import VisitDetail as VisitDetail
from gen_epix.omopdb.domain.model.omop import VisitOccurrence as VisitOccurrence
from gen_epix.omopdb.domain.model.omop import Vocabulary as Vocabulary
from gen_epix.omopdb.domain.model.omop.non_persistable import Subject as Subject

# List up model classes per service and sorted according to links topology
SORTED_MODELS_BY_SERVICE_TYPE: dict[enum.ServiceType, list[type[fastapp.Model]]] = (
    {  # pyright: ignore[reportAssignmentType]
        # Common models
        enum.ServiceType.AUTH: list(
            _COMMON_SORTED_MODELS_BY_SERVICE_TYPE[common_enum.ServiceType.AUTH]
        ),
        enum.ServiceType.SYSTEM: list(
            _COMMON_SORTED_MODELS_BY_SERVICE_TYPE[common_enum.ServiceType.SYSTEM]
        ),
        enum.ServiceType.RBAC: list(
            _COMMON_SORTED_MODELS_BY_SERVICE_TYPE[common_enum.ServiceType.RBAC]
        ),
        enum.ServiceType.ORGANIZATION: list(
            _COMMON_SORTED_MODELS_BY_SERVICE_TYPE[common_enum.ServiceType.ORGANIZATION]
        ),
        # Specific models
        enum.ServiceType.ABAC: list(
            _COMMON_SORTED_MODELS_BY_SERVICE_TYPE[common_enum.ServiceType.ABAC]
        )
        + [],
        enum.ServiceType.OMOP: [
            # Ordered topologically based on foreign key dependencies
            # Foundation tables (no dependencies)
            Location,
            CohortDefinition,
            Cohort,
            CdmSource,
            # Vocabulary system (circular dependencies resolved logically)
            Vocabulary,
            Domain,
            ConceptClass,
            Concept,
            Relationship,
            # Concept relationships (depend on concept and relationship)
            ConceptRelationship,
            ConceptAncestor,
            ConceptSynonym,
            # Reference/mapping tables
            DrugStrength,
            SourceToConceptMap,
            Metadata,
            # Care infrastructure
            CareSite,
            Provider,
            # Person and observations
            Person,
            ObservationPeriod,
            PayerPlanPeriod,
            # Visits (depend on person, care_site, provider, concept)
            VisitOccurrence,
            VisitDetail,
            # Clinical events (depend on person, visits, providers, concepts)
            ConditionOccurrence,
            ProcedureOccurrence,
            DrugExposure,
            DeviceExposure,
            Measurement,
            Observation,
            Specimen,
            Note,
            # Other tables depending at least on person
            ConditionEra,
            DrugEra,
            DoseEra,
            NoteNlp,
            Cost,
            LocationHistory,
            SurveyConduct,
            # General relationships
            FactRelationship,
            MeasurementRelation,
        ],
    }
)

SORTED_SERVICE_TYPES = tuple(SORTED_MODELS_BY_SERVICE_TYPE.keys())

COMMON_MODEL_MAP: dict[type[fastapp.Model], type[fastapp.Model]] = {}



from typing import ClassVar
from uuid import UUID

from pydantic import Field

import gen_epix.casedb.domain.model.geo as model
from gen_epix.commondb.domain.command import Command, CrudCommand

# Non-CRUD


class RetrieveContainingRegionCommand(Command):
    """
    Retrieve the regions that contain the specified regions.
    """

    region_ids: list[UUID] = Field(
        description="The IDs of the regions to retrieve containing regions for."
    )
    region_set_id: UUID = Field(
        description="The ID of the region set to that containing regions must belong to."
    )
    level: int = Field(description="The level of the region to retrieve.")


# CRUD


class RegionSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RegionSet


class RegionCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Region


class RegionRelationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RegionRelation


class RegionSetShapeCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.RegionSetShape



from collections.abc import Callable
from functools import partial

from pydantic import BaseModel


class Key:
    DEFAULT_KEY_GENERATOR_SEPARATOR = "|"

    def __init__(
        self,
        field_names: str | tuple[str, ...],
        key_generator: Callable[[BaseModel], str] | None = None,
    ):
        self._field_names: tuple[str, ...] = (
            (field_names,) if isinstance(field_names, str) else field_names
        )

        def _key_generator(field_names: tuple[str, ...], obj: BaseModel) -> str:
            return Key.DEFAULT_KEY_GENERATOR_SEPARATOR.join(
                f"{obj.__dict__[x]}" for x in field_names
            )

        self._key_generator: Callable[[BaseModel], str] = (
            key_generator
            if key_generator
            else partial(_key_generator, self._field_names)
        )

    @property
    def field_names(self) -> tuple[str, ...]:
        return self._field_names

    @property
    def key_generator(self) -> Callable[[BaseModel], str]:
        """
        Get the key generator callable.

        Returns
        -------
        Callable[[BaseModel], str]
            The callable used to generate the key from a BaseModel obj.
        """
        return self._key_generator

    def __call__(self, obj: BaseModel) -> str:
        return self._key_generator(obj)



# pylint: disable=useless-import-alias
from gen_epix.commondb.domain.repository import BaseAbacRepository as BaseAbacRepository
from gen_epix.commondb.domain.repository import (
    BaseOrganizationRepository as BaseOrganizationRepository,
)
from gen_epix.commondb.domain.repository import (
    BaseSystemRepository as BaseSystemRepository,
)
from gen_epix.seqdb.domain.repository.file import (
    BaseFileRepository as BaseFileRepository,
)
from gen_epix.seqdb.domain.repository.seq import BaseSeqRepository as BaseSeqRepository



from collections.abc import Callable
from typing import Any, NoReturn
from uuid import UUID

from fastapi import APIRouter, FastAPI
from pydantic import BaseModel as PydanticBaseModel

from gen_epix.casedb.domain import command, enum, model
from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.fastapp import App
from gen_epix.fastapp.api.crud_endpoint_generator import CrudEndpointGenerator


class UpdateDiseaseEtiologicalAgentRequestBody(PydanticBaseModel):
    etiologies: list[model.Etiology]


def create_ontology_endpoints(
    router: APIRouter | FastAPI,
    app: App,
    handle_exception: Callable[[str, Any, Exception], NoReturn] | None = None,
    **kwargs: Any,
) -> None:
    assert handle_exception
    app_impl: AppImplDetails = app.impl
    registered_user_dependency = app_impl.registered_user_dependency

    @router.put(
        "/diseases/{disease_id}/etiological_agents",
        operation_id="diseases__put__etiological_agents",
        name="Disease_EtiologicalAgent",
        description=command.DiseaseEtiologicalAgentUpdateAssociationCommand.__doc__,
    )
    async def diseases__put__concepts(
        user: registered_user_dependency,  # type: ignore
        disease_id: UUID,
        request_body: UpdateDiseaseEtiologicalAgentRequestBody,
    ) -> list[model.Etiology]:
        try:
            cmd = command.DiseaseEtiologicalAgentUpdateAssociationCommand(
                user=user,
                obj_id1=disease_id,
                association_objs=request_body.etiologies,
                props={"return_id": False},
            )
            retval: list[model.Etiology] = app.handle(cmd)
        except Exception as exception:
            handle_exception("d5459ee4", user, exception)
        return retval

    # CRUD
    crud_endpoint_sets = CrudEndpointGenerator.create_crud_endpoint_set_for_domain(
        app,
        service_type=enum.ServiceType.ONTOLOGY,
        user_dependency=registered_user_dependency,
    )
    CrudEndpointGenerator.generate_endpoints(
        router, crud_endpoint_sets, handle_exception
    )



from gen_epix.commondb.app_setup import create_fast_api
from gen_epix.commondb.config import AppCfg
from gen_epix.commondb.util import get_package_version
from gen_epix.omopdb.api.router import create_routers
from gen_epix.omopdb.domain import enum
from gen_epix.omopdb.env import AppComposer

APP_NAME = "OMOPDB"

# Data for OpenAPI schema
SCHEMA_KWARGS = {
    "title": "Gen-EpiX omopdb",
    "summary": "Genomic Epidemiology platform for disease X, omopdb app",
    "description": "The omopdb app manages clinical and epidemiological data of persons or subjects of non-human origin.",
    "version": get_package_version(),
    "terms_of_service": "http://example.com/terms/",
    "contact": {
        "name": "RIVM CIb IDS bioinformatics group",
        "url": "https://github.com/RIVM-bioinformatics/gen-epix-api",
        "email": "ids-bioinformatics@rivm.nl",
    },
    "license_info": {
        "name": "License to be confirmed",
        "identifier": "Apache-2.0",
    },
}

# Get configuration data and environment
APP_CFG = AppCfg(APP_NAME, enum.ServiceType, enum.RepositoryType)
APP_COMPOSER = AppComposer(APP_CFG)

# Create fastapi
FAST_API = create_fast_api(
    APP_COMPOSER.app,
    create_routers_fn=create_routers,
    setup_logger=APP_CFG.setup_logger,
    api_logger=APP_CFG.api_logger,
    debug=APP_CFG.cfg.app.debug,
    update_openapi_schema=True,
    update_openapi_kwargs={
        "get_openapi_kwargs": SCHEMA_KWARGS,
        "fix_schema": True,
        "auth_service": APP_COMPOSER.services[enum.ServiceType.AUTH],
    },
)

# TODO: app variable added for backwards compatibility with startup code that imports "app". Remove once that code is updated as well.
app = FAST_API



# pylint: disable=useless-import-alias
from gen_epix.commondb.services.abac import AbacService as AbacService
from gen_epix.commondb.services.organization import (
    OrganizationService as OrganizationService,
)
from gen_epix.commondb.services.rbac import RbacService as RbacService
from gen_epix.commondb.services.system import SystemService as SystemService
from gen_epix.commondb.services.user_manager import UserManager as UserManager
from gen_epix.fastapp.services.auth import AuthService as AuthService



import threading

import sqlalchemy as sa
from sqlalchemy import Engine

DEFAULT_POOL_RECYCLE = 1800


class EngineFactory:
    """
    Static factory class to create and manage SQLAlchemy engine objs.
    """

    _LOCK = threading.Lock()
    _ENGINE_MAP: dict[tuple, Engine] = {}

    def __init__(self) -> None:
        raise ValueError(
            "EngineFactory is a static class and should not be instantiated."
        )

    @classmethod
    def create_engine(
        cls,
        connection_string: str,
        echo: bool = False,
        pool_recycle: int = DEFAULT_POOL_RECYCLE,
    ) -> Engine:
        """
        Create a new SQLAlchemy engine or return an existing one for the given connection string.

        Args:
            connection_string (str): The database connection string.
            echo (bool): If True, the engine will log all statements as well as a repr() of their parameter lists to the default log handler, which defaults to sys.stdout. Defaults to False.

        Returns:
            Engine: The SQLAlchemy engine obj.
        """
        key = cls._compose_key(connection_string, echo=echo, pool_recycle=pool_recycle)
        with cls._LOCK:
            if key not in cls._ENGINE_MAP:
                engine = sa.create_engine(
                    connection_string, echo=echo, pool_recycle=pool_recycle
                )
                cls._ENGINE_MAP[key] = engine
            return cls._ENGINE_MAP[key]

    @classmethod
    def _compose_key(
        cls,
        connection_string: str,
        echo: bool = False,
        pool_recycle: int = DEFAULT_POOL_RECYCLE,
    ) -> tuple[str, bool, int]:
        return (connection_string, echo, pool_recycle)



import re
from decimal import Decimal, InvalidOperation
from itertools import combinations
from typing import Any, NoReturn
from uuid import UUID

from gen_epix.casedb.domain import command, enum, model
from gen_epix.casedb.domain.enum import (
    CaseColDataRule,
    ColType,
    ColTypeSet,
    ConceptRelationType,
    RegionRelationType,
)
from gen_epix.casedb.services.case.base import BaseCaseService
from gen_epix.commondb.util import map_paired_elements
from gen_epix.fastapp import CrudOperation
from gen_epix.filter import UuidSetFilter
from gen_epix.filter.composite import CompositeFilter
from gen_epix.filter.enum import LogicalOperator
from gen_epix.transform import Transformer
from gen_epix.transform.adapter import ObjectAdapter
from gen_epix.transform.enum import NoMatchStrategy, TimeUnit, TimeUnitTransformStrategy
from gen_epix.transform.transform_result import TransformResult
from gen_epix.transform.transformers import IntervalTransformer
from gen_epix.transform.transformers.interval import IntervalToIntervalTransformer
from gen_epix.transform.transformers.iso_time import IsoTimeTransformer


class CaseTransformer(Transformer):
    N_DECIMALS = {
        ColType.DECIMAL_0: 0,
        ColType.DECIMAL_1: 1,
        ColType.DECIMAL_2: 2,
        ColType.DECIMAL_3: 3,
        ColType.DECIMAL_4: 4,
        ColType.DECIMAL_5: 5,
        ColType.DECIMAL_6: 6,
    }

    FLOAT_PATTERN = re.compile(
        r"^[+-]?("
        r"(?:\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?"  # normal floats + scientific notation
        r"|inf(?:inity)?"  # inf / infinity
        r"|nan"  # nan
        r")$",
        re.IGNORECASE,
    )

    TIME_YEAR_PATTERN = re.compile(r"^\d{4}$")
    TIME_QUARTER_PATTERN = re.compile(r"^\d{4}-Q[1-4]$")
    TIME_MONTH_PATTERN = re.compile(r"^\d{4}-(0[1-9]|1[0-2])$")
    TIME_WEEK_PATTERN = re.compile(r"^\d{4}-W(0[1-9]|[1-4]\d|5[0-3])$")
    TIME_DAY_PATTERN = re.compile(r"^\d{4}-(0[1-9]|1[0-2])-(0[1-9]|[12]\d|3[01])$")

    TIME_MATCHERS = {
        ColType.TIME_YEAR: lambda x: (
            x if x is None or CaseTransformer.TIME_YEAR_PATTERN.match(x) else NoReturn
        ),
        ColType.TIME_QUARTER: lambda x: (
            x
            if x is None or CaseTransformer.TIME_QUARTER_PATTERN.match(x)
            else NoReturn
        ),
        ColType.TIME_MONTH: lambda x: (
            x if x is None or CaseTransformer.TIME_MONTH_PATTERN.match(x) else NoReturn
        ),
        ColType.TIME_WEEK: lambda x: (
            x if x is None or CaseTransformer.TIME_WEEK_PATTERN.match(x) else NoReturn
        ),
        ColType.TIME_DAY: lambda x: (
            x if x is None or CaseTransformer.TIME_DAY_PATTERN.match(x) else NoReturn
        ),
    }

    COL_TYPE_TO_TIME_UNIT = {
        ColType.TIME_YEAR: TimeUnit.YEAR,
        ColType.TIME_QUARTER: TimeUnit.QUARTER,
        ColType.TIME_MONTH: TimeUnit.MONTH,
        ColType.TIME_WEEK: TimeUnit.WEEK,
        ColType.TIME_DAY: TimeUnit.DAY,
    }

    @staticmethod
    def _transform_decimal(value: str | None, n_decimals: int) -> str | None | NoReturn:
        if value is None:
            return None
        if re.match(r"[+-]?([0-9]*[.,])?[0-9]+", value) is None:
            return NoReturn
        try:
            num_value = round(Decimal(value), n_decimals)
            return str(num_value)
        except (ValueError, InvalidOperation):
            return NoReturn

    def __init__(
        self,
        case_service: BaseCaseService,
        complete_case_type: model.CompleteCaseType,
    ):
        self.case_service = case_service
        self.complete_case_type = complete_case_type
        # Unique concept and region sets across the complete case type
        self.concept_set_ids: set[UUID] = set()
        self.interval_concept_set_ids: set[UUID] = set()
        self.regex_concept_set_ids: set[UUID] = set()
        self.region_set_ids: set[UUID] = set()
        # dict[concept_set_id, dict[lower(str(concept_id)|abbrevation|name), concept_id]]
        self.concept_value_maps: dict[UUID, dict[str, str]] = {}
        # dict[(from_concept_set_id, to_concept_set_id), dict[from_concept_id, to_concept_id]]
        self.concept_relation_maps: dict[tuple[UUID, UUID], dict[str, str]] = {}
        # dict[to_concept_set_id, IntervalTransformer]
        self.interval_transformers: dict[UUID, IntervalTransformer] = {}
        # dict[concept_set_id, pattern_matcher]
        self.regex_patterns: dict[UUID, re.Pattern] = {}
        # dict[region_set_id, dict[lower(str(region_id)|code|name), region_id]]
        self.region_value_maps: dict[UUID, dict[str, str]] = {}
        # dict[(from_region_set_id, to_region_set_id), dict[from_region_id, to_region_id]]
        self.region_relation_maps: dict[tuple[UUID, UUID], dict[str, str]] = {}
        # dict[from_region_id, to_region_id] - flattened mapping across all region sets
        self.region_contained_in: dict[UUID, UUID] = {}
        # dict[region_id, region_object] - for looking up region details
        self.regions: dict[UUID, model.Region] = {}
        # dict[lower(str(organization_id)|name|legal_entity_code), concept_id]
        self.organization_value_map: dict[str, str] = {}

        self._init_metadata()

    def transform(self, obj: ObjectAdapter) -> ObjectAdapter:
        raise NotImplementedError()

    def __call__(self, obj: Any) -> TransformResult:
        case_validation_report, contents, updated_contents, is_update = (
            self._setup_reporting(obj)
        )

        self._handle_unknown_columns(case_validation_report, contents)

        self._transform_individual_values(
            case_validation_report, contents, updated_contents, is_update
        )

        self._transform_value_pairs(case_validation_report, contents, updated_contents)

        return TransformResult(
            success=True, original_object=obj, transformed_object=case_validation_report
        )

    def _setup_reporting(self, obj: command.ValidateCasesCommand) -> tuple[
        model.CaseValidationReport,
        list[dict[UUID, str | None]],
        list[dict[UUID, str | None]],
        bool,
    ]:
        """Set up the case validation report and content lists for processing."""
        if obj.case_type_id != self.complete_case_type.id:
            raise ValueError("Invalid case type")

        is_update = obj.is_update
        contents = [case.content for case in obj.cases]

        # Create case validation report with empty content for cases
        case_validation_report = model.CaseValidationReport(
            case_type_id=obj.case_type_id,
            created_in_data_collection_id=obj.created_in_data_collection_id,
            is_update=obj.is_update,
            data_collection_ids=obj.data_collection_ids,
            validated_cases=[
                model.ValidatedCase(
                    case=model.CaseForCreateUpdate(
                        **x.model_dump(exclude={"content"}), content={}
                    ),
                    data_issues=[],
                )
                for x in obj.cases
            ],
        )
        updated_contents = [
            x.case.content for x in case_validation_report.validated_cases
        ]

        return case_validation_report, contents, updated_contents, is_update

    def _handle_unknown_columns(
        self,
        case_validation_report: model.CaseValidationReport,
        contents: list[dict[UUID, str | None]],
    ) -> None:
        """Handle any unknown case type columns and add appropriate data issues."""
        for i, content in enumerate(contents):
            for case_type_col_id in content.keys():
                if case_type_col_id in self.complete_case_type.case_type_cols:
                    continue
                # Unknown case type col
                case_validation_report.validated_cases[i].data_issues.append(
                    model.CaseDataIssue(
                        case_type_col_id=case_type_col_id,
                        original_value=content[case_type_col_id],
                        updated_value=None,
                        data_rule=CaseColDataRule.UNAUTHORIZED,
                        details="Unknown case type column",
                    )
                )

    def _transform_individual_values(
        self,
        case_validation_report: model.CaseValidationReport,
        contents: list[dict[UUID, str | None]],
        updated_contents: list[dict[UUID, str | None]],
        is_update: bool,
    ) -> None:
        """Validate and transform individual values."""
        msg_template = "{orig_value}"
        for case_type_col in self.complete_case_type.case_type_cols.values():
            case_type_col_id = case_type_col.id
            assert case_type_col_id is not None
            col = self.complete_case_type.cols[case_type_col.col_id]
            if col.col_type == ColType.REGULAR_LANGUAGE:
                assert col.concept_set_id
                transform_fn = lambda x: (
                    x
                    if x is None or self.regex_patterns[col.concept_set_id].match(x)
                    else NoReturn
                )
                msg_template = "{orig_value} does not match regex"
            elif col.col_type in ColTypeSet.STRING_SET.value:
                assert col.concept_set_id
                concept_value_map = self.concept_value_maps[col.concept_set_id]
                transform_fn = lambda x: (
                    concept_value_map.get(x.lower(), NoReturn) if x else None
                )
                msg_template = "{orig_value} cannot be mapped to concept"
            elif col.col_type in ColTypeSet.HAS_REGION_SET.value:
                assert col.region_set_id
                region_value_map = self.region_value_maps[col.region_set_id]
                transform_fn = lambda x: (
                    region_value_map.get(x.lower(), NoReturn) if x else None
                )
                msg_template = "{orig_value} cannot be mapped to region"
            elif col.col_type in ColTypeSet.TIME.value:
                transform_fn = self.TIME_MATCHERS[col.col_type]
                msg_template = (
                    "{orig_value} is not a valid " + col.col_type.value + " value"
                )
            elif col.col_type in ColTypeSet.NUMBER.value:
                n_decimals = self.N_DECIMALS[col.col_type]
                transform_fn = lambda x: CaseTransformer._transform_decimal(
                    x, n_decimals
                )
            elif col.col_type == ColType.ORGANIZATION:
                organization_value_map = self.organization_value_map
                transform_fn = lambda x: (
                    organization_value_map.get(x.lower(), NoReturn) if x else None
                )
                msg_template = "{orig_value} cannot be mapped to organization"
            else:
                # TODO: transform any other col_types
                transform_fn = lambda x: x
            # Update value
            for i, (content, updated_content) in enumerate(
                zip(contents, updated_contents)
            ):
                if case_type_col_id not in content:
                    continue
                orig_value = content[case_type_col_id]
                if orig_value is None:
                    if is_update:
                        # Deletion of value by providing None -> keep in content
                        updated_content[case_type_col_id] = None
                    # Unnecessary None -> do not add
                    continue
                new_value = transform_fn(orig_value)
                if new_value == NoReturn:
                    new_value = None
                    # No mapping found
                    case_validation_report.validated_cases[i].data_issues.append(
                        model.CaseDataIssue(
                            case_type_col_id=case_type_col_id,
                            original_value=orig_value,
                            updated_value=new_value,
                            data_rule=CaseColDataRule.INVALID,
                            details=msg_template.format(orig_value=orig_value),
                        )
                    )
                    continue
                updated_content[case_type_col_id] = new_value

    def _transform_value_pairs(
        self,
        case_validation_report: model.CaseValidationReport,
        contents: list[dict[UUID, str | None]],
        updated_contents: list[dict[UUID, str | None]],
    ) -> None:
        """
        Validate and transform pairs of values.
        This method assumes that only standard values are present in updated_contents.
        """

        # Go over each case type dimension, i.e. (Dim, occurrence in the CaseType) combination
        for case_type_dim in self.complete_case_type.case_type_dims:
            dim_type = self.complete_case_type.dims[case_type_dim.dim_id]
            case_type_col_ids = case_type_dim.case_type_col_order
            # Generate all ordered pairs of case type col IDs in this case type dimension, in both directions
            # TODO: do not generate the pairs for types of dimensions that are not handled
            col_pairs = list(combinations(case_type_col_ids, 2)) + list(
                combinations(case_type_col_ids[::-1], 2)
            )
            # Handle each type of dimension
            if dim_type.dim_type == enum.DimType.GEO:
                self._transform_geo_value_pairs(
                    col_pairs, case_validation_report, contents, updated_contents
                )
            elif dim_type.dim_type == enum.DimType.TIME:
                self._transform_time_value_pairs(
                    col_pairs, case_validation_report, contents, updated_contents
                )
            elif dim_type.dim_type == enum.DimType.NUMBER:
                self._transform_number_value_pairs(
                    col_pairs, case_validation_report, contents, updated_contents
                )
            elif dim_type.dim_type == enum.DimType.TEXT:
                pass
            else:
                continue

    def _transform_geo_value_pairs(
        self,
        col_pairs: list[tuple[UUID, UUID]],
        case_validation_report: model.CaseValidationReport,
        contents: list[dict[UUID, str | None]],
        updated_contents: list[dict[UUID, str | None]],
    ) -> None:
        """
        Validate and transform GEO pairs of values.
        Applies only to GEO_REGION-GEO_REGION column pairs.
        """
        for col_pair in col_pairs:
            col1 = self.complete_case_type.cols[
                self.complete_case_type.case_type_cols[col_pair[0]].col_id
            ]
            col2 = self.complete_case_type.cols[
                self.complete_case_type.case_type_cols[col_pair[1]].col_id
            ]
            if (
                col1.col_type != ColType.GEO_REGION
                or col2.col_type != ColType.GEO_REGION
            ):
                continue
            assert col1.region_set_id is not None
            assert col2.region_set_id is not None
            region_relation_map = self.region_relation_maps.get(
                (col1.region_set_id, col2.region_set_id), {}
            )
            for i, (content, updated_content) in enumerate(
                zip(contents, updated_contents)
            ):
                from_region_id = updated_content.get(col_pair[0])
                if from_region_id is None:
                    continue
                to_region_id = region_relation_map.get(from_region_id)
                if to_region_id is None:
                    continue

                self._set_derived_value(
                    case_validation_report,
                    i,
                    content,
                    updated_content,
                    col_pair,
                    to_region_id,
                )

    def _transform_time_value_pairs(
        self,
        col_pairs: list[tuple[UUID, UUID]],
        case_validation_report: model.CaseValidationReport,
        contents: list[dict[UUID, str | None]],
        updated_contents: list[dict[UUID, str | None]],
    ) -> None:
        """Validate and transform TIME pairs of values."""
        # For TIME dim: use IsoTimeTransformer with TimeUnitTransformStrategy.EXACT_ONLY to transform from-values to to-values. When no transformation is possible (e.g. from MONTH to DAY), skip that pair of cols to avoid a call to the IsoTimeTransformer.
        for col_pair in col_pairs:
            col1 = self.complete_case_type.cols[
                self.complete_case_type.case_type_cols[col_pair[0]].col_id
            ]
            col2 = self.complete_case_type.cols[
                self.complete_case_type.case_type_cols[col_pair[1]].col_id
            ]
            # Skip if columns are not TIME types
            if (
                col1.col_type not in ColTypeSet.TIME.value
                or col2.col_type not in ColTypeSet.TIME.value
            ):
                continue

            from_time_unit = self.COL_TYPE_TO_TIME_UNIT[col1.col_type]
            to_time_unit = self.COL_TYPE_TO_TIME_UNIT[col2.col_type]

            # Skip if transformation is not possible
            if not IsoTimeTransformer.can_transform_time(from_time_unit, to_time_unit):
                continue

            # Create IsoTimeTransformer for this transformation
            time_transformer = IsoTimeTransformer(
                field_name="time_value",
                src_unit=from_time_unit,
                tgt_unit=to_time_unit,
                strategy=TimeUnitTransformStrategy.EXACT_ONLY,
            )

            for i, (content, updated_content) in enumerate(
                zip(contents, updated_contents)
            ):
                from_time = updated_content.get(col_pair[0])
                if from_time is None:
                    continue
                col1 = self.complete_case_type.cols[
                    self.complete_case_type.case_type_cols[col_pair[0]].col_id
                ]

                # Check if from_time is a valid time value for col1's type
                if self.TIME_MATCHERS[col1.col_type](from_time) is NoReturn:
                    continue

                # Transform the time value using ObjectAdapter
                try:
                    adapter = ObjectAdapter({"time_value": from_time})
                    transformed_adapter = time_transformer.transform(adapter)
                    to_time = transformed_adapter.get("time_value")
                except Exception:
                    # Skip if transformation fails
                    continue
                if to_time is None:
                    continue

                self._set_derived_value(
                    case_validation_report,
                    i,
                    content,
                    updated_content,
                    col_pair,
                    to_time,
                )

    def _transform_number_value_pairs(
        self,
        col_pairs: list[tuple[UUID, UUID]],
        case_validation_report: model.CaseValidationReport,
        contents: list[dict[UUID, str | None]],
        updated_contents: list[dict[UUID, str | None]],
    ) -> None:
        """
        Validate and transform NUMBER pairs of values.
        Applies only to DECIMAL_XXX-INTERVAL and INTERVAL-INTERVAL column pairs.
        """
        for col_pair in col_pairs:
            col1 = self.complete_case_type.cols[
                self.complete_case_type.case_type_cols[col_pair[0]].col_id
            ]
            col2 = self.complete_case_type.cols[
                self.complete_case_type.case_type_cols[col_pair[1]].col_id
            ]

            # Only handle DECIMAL_XXX-INTERVAL and INTERVAL-INTERVAL pairs
            is_col1_decimal = col1.col_type in ColTypeSet.NUMBER.value
            is_col2_interval = col2.col_type == ColType.INTERVAL
            is_col1_interval = col1.col_type == ColType.INTERVAL

            # DECIMAL_XXX -> INTERVAL transformation
            if is_col1_decimal and is_col2_interval:
                self._transform_decimal_to_interval(
                    col_pair,
                    col2,
                    case_validation_report,
                    contents,
                    updated_contents,
                )

            # INTERVAL -> INTERVAL transformation using IntervalToIntervalTransformer
            elif is_col1_interval and is_col2_interval:
                self._transform_interval_to_interval(
                    col_pair,
                    col1,
                    col2,
                    case_validation_report,
                    contents,
                    updated_contents,
                )

    def _transform_decimal_to_interval(
        self,
        col_pair: tuple[UUID, UUID],
        col2: model.Col,
        case_validation_report: model.CaseValidationReport,
        contents: list[dict[UUID, str | None]],
        updated_contents: list[dict[UUID, str | None]],
    ) -> None:
        """Process DECIMAL_XXX -> INTERVAL transformation."""
        # Check if we have an interval transformer for col2's concept set
        if col2.concept_set_id is None:
            return
        interval_transformer = self.interval_transformers.get(col2.concept_set_id)
        if interval_transformer is None:
            return

        for i, (content, updated_content) in enumerate(zip(contents, updated_contents)):
            from_number_str = updated_content.get(col_pair[0])
            if from_number_str is None:
                continue
            from_number = float(from_number_str)

            if not interval_transformer.is_transformable(from_number):
                continue

            # Use the interval transformer with "value" field name
            adapter = ObjectAdapter({"value": from_number})
            transformed_adapter = interval_transformer.transform(adapter)
            to_interval_id = transformed_adapter.get("value")
            if to_interval_id is None:
                continue

            self._set_derived_value(
                case_validation_report,
                i,
                content,
                updated_content,
                col_pair,
                to_interval_id,
            )

    def _transform_interval_to_interval(
        self,
        col_pair: tuple[UUID, UUID],
        col1: model.Col,
        col2: model.Col,
        case_validation_report: model.CaseValidationReport,
        contents: list[dict[UUID, str | None]],
        updated_contents: list[dict[UUID, str | None]],
    ) -> None:
        """Process INTERVAL -> INTERVAL transformation using IntervalToIntervalTransformer."""
        # Check if both columns have concept sets
        if col1.concept_set_id is None or col2.concept_set_id is None:
            return

        # Skip if trying to map to the same concept set
        if col1.concept_set_id == col2.concept_set_id:
            return

        # Get interval transformers for both concept sets
        src_transformer = self.interval_transformers.get(col1.concept_set_id)
        tgt_transformer = self.interval_transformers.get(col2.concept_set_id)

        if src_transformer is None or tgt_transformer is None:
            return

        # Create IntervalToIntervalTransformer
        try:
            interval_to_interval_transformer = IntervalToIntervalTransformer(
                src_field="interval_value",
                src_interval_names=src_transformer._interval_names,
                src_lower_bounds=src_transformer._lower_bounds,
                src_upper_bounds=src_transformer._upper_bounds,
                tgt_interval_names=tgt_transformer._interval_names,
                tgt_lower_bounds=tgt_transformer._lower_bounds,
                tgt_upper_bounds=tgt_transformer._upper_bounds,
                src_lower_bound_is_inclusive=src_transformer._lower_bound_is_inclusive,
                src_upper_bound_is_inclusive=src_transformer._upper_bound_is_inclusive,
                tgt_lower_bound_is_inclusive=tgt_transformer._lower_bound_is_inclusive,
                tgt_upper_bound_is_inclusive=tgt_transformer._upper_bound_is_inclusive,
                overlap_strategy="largest_overlap",
                no_match_strategy=NoMatchStrategy.SET_NONE,
            )
        except Exception:
            # Skip if transformer creation fails
            return

        for i, (content, updated_content) in enumerate(zip(contents, updated_contents)):
            from_interval_id = updated_content.get(col_pair[0])
            if from_interval_id is None:
                continue

            # TODO: replace by pre-calculated interval_relation_map for efficiency
            # Check if transformation is possible
            if not interval_to_interval_transformer.is_transformable(from_interval_id):
                continue
            # Transform the interval value using ObjectAdapter
            try:
                adapter = ObjectAdapter({"interval_value": from_interval_id})
                transformed_adapter = interval_to_interval_transformer.transform(
                    adapter
                )
                to_interval_id = transformed_adapter.get("interval_value")
            except Exception:
                # Skip if transformation fails
                continue
            if to_interval_id is None:
                continue

            self._set_derived_value(
                case_validation_report,
                i,
                content,
                updated_content,
                col_pair,
                to_interval_id,
            )

    def _set_derived_value(
        self,
        case_validation_report: model.CaseValidationReport,
        case_index: int,
        content: dict,
        updated_content: dict,
        col_pair: tuple,
        new_value: str,
    ) -> None:
        # Add derived value to updated_content
        updated_content[col_pair[1]] = new_value

        # Log data issue in validation report
        if (
            col_pair[1] in updated_content
            and updated_content[col_pair[1]] is not None
            and updated_content[col_pair[1]] != new_value
        ):
            # Overwrite existing different value
            case_validation_report.validated_cases[case_index].data_issues.append(
                model.CaseDataIssue(
                    case_type_col_id=col_pair[1],
                    original_value=content[col_pair[1]],
                    updated_value=new_value,
                    data_rule=CaseColDataRule.CONFLICT,
                    details=f"Value overwritten based on derived value from '{content[col_pair[0]]}'",
                )
            )
        else:
            # Set derived value
            case_validation_report.validated_cases[case_index].data_issues.append(
                model.CaseDataIssue(
                    case_type_col_id=col_pair[1],
                    original_value=content.get(col_pair[1]),
                    updated_value=new_value,
                    data_rule=CaseColDataRule.DERIVED,
                    # TODO: time dimension should always derive from the highest resolution value first
                    details=f"Value derived from '{content.get(col_pair[0])}'",
                )
            )

    def _init_metadata(self) -> None:
        self._init_set_metadata()
        self._init_concept_metadata()
        self._init_region_metadata()
        self._init_organization_metadata()

    def _init_set_metadata(self) -> None:
        self.concept_set_ids = set()
        self.interval_concept_set_ids = set()
        self.regex_concept_set_ids = set()
        self.region_set_ids = set()
        # Get unique concept and region sets across the complete case type
        for col in self.complete_case_type.cols.values():
            if col.col_type in ColTypeSet.HAS_CONCEPT_SET.value:
                assert col.concept_set_id
                self.concept_set_ids.add(col.concept_set_id)
                if col.col_type == ColType.INTERVAL:
                    self.interval_concept_set_ids.add(col.concept_set_id)
                elif col.col_type == ColType.REGULAR_LANGUAGE:
                    self.regex_concept_set_ids.add(col.concept_set_id)
            elif col.col_type in ColTypeSet.HAS_REGION_SET.value:
                assert col.region_set_id
                self.region_set_ids.add(col.region_set_id)

    def _init_concept_metadata(self) -> None:
        self.concept_value_maps = {}
        self.concept_relation_maps = {}
        self.interval_transformers = {}
        self.regex_patterns = {}

        # Retrieve relevant concept sets, concepts, and relations
        concept_sets, concept_set_concepts_map, concepts, concept_contained_in = (
            self._retrieve_concept_data()
        )

        # Fill concept_value_maps
        for concept_set_id, concept_ids in concept_set_concepts_map.items():
            self.concept_value_maps[concept_set_id] = (
                {str(x).lower(): str(x) for x in concept_ids}
                | {
                    concepts[x].code.lower(): str(x)
                    for x in concept_ids
                    if concepts[x].code is not None
                }
                | {
                    concepts[x].name.lower(): str(x)
                    for x in concept_ids
                    if concepts[x].name is not None
                }
            )

        # Fill in concept relation maps
        for (
            from_concept_set_id,
            to_concept_set_id,
        ), mapping in concept_contained_in.items():
            self.concept_relation_maps.setdefault(
                (from_concept_set_id, to_concept_set_id), {}
            )
            self.concept_relation_maps[(from_concept_set_id, to_concept_set_id)].update(
                mapping
            )

        # Fill in interval mappers
        for concept_set_id in self.interval_concept_set_ids:
            interval_concepts = [
                concepts[x] for x in concept_set_concepts_map[concept_set_id]
            ]
            self.interval_transformers[concept_set_id] = IntervalTransformer(
                "value",  # src_field should match the field name used in ObjectAdapter
                [str(x.id) for x in interval_concepts],
                [x.props["lb"] for x in interval_concepts],
                [x.props["ub"] for x in interval_concepts],
                lower_bound_is_inclusive=[x.props["lb_in"] for x in interval_concepts],
                upper_bound_is_inclusive=[x.props["ub_in"] for x in interval_concepts],
            )

        # Fill in regex matchers
        for concept_set_id in self.regex_concept_set_ids:
            concept_set = concept_sets[concept_set_id]
            assert concept_set.regex is not None
            self.regex_patterns[concept_set_id] = re.compile(concept_set.regex)

    def _init_region_metadata(self) -> None:
        self.region_value_maps = {}
        self.region_relation_maps = {}
        self.region_contained_in = {}
        self.regions = {}

        # Retrieve relevant regions and relations
        regions, region_set_regions_map, region_contained_in = (
            self._retrieve_region_data()
        )

        # Store regions for lookup
        self.regions = regions

        # Fill in region value maps
        for region_set_id in self.region_set_ids:
            curr_regions = [regions[x] for x in region_set_regions_map[region_set_id]]
            self.region_value_maps[region_set_id] = (  # type:ignore[arg-type]
                {
                    str(x.id).lower(): str(x.id) for x in curr_regions
                }  # type:ignore[assignment]
                | {x.code.lower(): str(x.id) for x in curr_regions if x.code}
                | {x.name.lower(): str(x.id) for x in curr_regions if x.name}
            )

        # Fill in region relation maps
        for (
            from_region_set_id,
            to_region_set_id,
        ), mapping in region_contained_in.items():
            self.region_relation_maps.setdefault(
                (from_region_set_id, to_region_set_id), {}
            )
            self.region_relation_maps[(from_region_set_id, to_region_set_id)].update(
                mapping
            )
            self.region_contained_in.update(mapping)

    def _init_organization_metadata(self) -> None:
        self.organization_value_map = {}

        # Retrieve organizations
        organizations = self._retrieve_organization_data()

        # Fill in organization value map
        self.organization_value_map = (
            {str(x.id).lower(): str(x.id) for x in organizations}
            | {
                x.legal_entity_code.lower(): str(x.id)
                for x in organizations
                if x.legal_entity_code
            }
            | {x.name.lower(): str(x.id) for x in organizations if x.name}
        )

    def _retrieve_concept_data(
        self,
    ) -> tuple[
        dict[UUID, model.ConceptSet],
        dict[UUID, set[UUID]],
        dict[UUID, model.Concept],
        dict[tuple[UUID, UUID], dict[str, str]],
    ]:
        app = self.case_service.app
        # Retrieve relevant concepts sets
        concept_sets: dict[UUID, model.ConceptSet] = {
            x.id: x
            for x in app.handle(
                command.ConceptSetCrudCommand(
                    operation=CrudOperation.READ_ALL,
                    query_filter=UuidSetFilter(
                        key="id", members=frozenset(self.concept_set_ids)
                    ),
                )
            )
        }
        # Retrieve relevant concepts
        concepts: dict[UUID, model.Concept] = {
            x.id: x
            for x in app.handle(
                command.ConceptCrudCommand(
                    operation=CrudOperation.READ_ALL,
                    query_filter=UuidSetFilter(
                        key="concept_set_id", members=frozenset(self.concept_set_ids)
                    ),
                )
            )
        }
        # Map concepts to sets
        concept_set_concepts_map: dict[UUID, set[UUID]] = (
            map_paired_elements(  # type:ignore[assignment]
                [(x.concept_set_id, x.id) for x in concepts.values()],
                as_set=True,
            )
        )

        # concept_contained_in maps (from_concept_set_id, to_concept_set_id) to map from_concept_id to to_concept_id
        concept_contained_in: dict[tuple[UUID, UUID], dict[str, str]] = {}
        for concept_relation in app.handle(
            command.ConceptRelationCrudCommand(
                operation=CrudOperation.READ_ALL,
                query_filter=CompositeFilter(
                    filters=[
                        UuidSetFilter(
                            key="from_concept_id",
                            members=frozenset(concepts.keys()),
                        ),
                        UuidSetFilter(
                            key="to_concept_id",
                            members=frozenset(concepts.keys()),
                        ),
                    ],
                    operator=LogicalOperator.AND,
                ),
            )
        ):
            from_concept = concepts[concept_relation.from_concept_id]
            to_concept = concepts[concept_relation.to_concept_id]
            if concept_relation.relation == ConceptRelationType.CONTAINS:
                # Swap concepts to convert from contains to is-contained-in
                from_concept, to_concept = to_concept, from_concept
            else:
                # Only contains and is-contained-in relationships considered
                continue
            key = (from_concept.concept_set_id, to_concept.concept_set_id)
            concept_contained_in.setdefault(key, {})
            assert from_concept.id is not None
            assert to_concept.id is not None
            concept_contained_in[key][str(from_concept.id)] = str(to_concept.id)

        return concept_sets, concept_set_concepts_map, concepts, concept_contained_in

    def _retrieve_region_data(self) -> tuple[
        dict[UUID, model.Region],
        dict[UUID, set[UUID]],
        dict[tuple[UUID, UUID], dict[str, str]],
    ]:
        app = self.case_service.app
        # Retrieve relevant regions
        regions: dict[UUID, model.Region] = {
            x.id: x
            for x in app.handle(
                command.RegionCrudCommand(
                    operation=CrudOperation.READ_ALL,
                    query_filter=UuidSetFilter(
                        key="region_set_id", members=frozenset(self.region_set_ids)
                    ),
                )
            )
        }
        # Map regions to sets
        region_set_regions_map: dict[UUID, set[UUID]] = (
            map_paired_elements(  # type:ignore[assignment]
                [(x.region_set_id, x.id) for x in regions.values()],
                as_set=True,
            )
        )

        # Retrieve region relations
        region_contained_in: dict[tuple[UUID, UUID], dict[UUID, UUID]] = {}
        region_relations = app.handle(
            command.RegionRelationCrudCommand(
                operation=CrudOperation.READ_ALL,
            )
        )
        region_relations = [
            x
            for x in region_relations
            if x.from_region_id in regions and x.to_region_id in regions
        ]
        for region_relation in region_relations:
            from_region = regions[region_relation.from_region_id]
            to_region = regions[region_relation.to_region_id]
            if region_relation.relation == RegionRelationType.CONTAINS:
                # Swap regions to convert from contains to is-contained-in
                from_region, to_region = to_region, from_region
            else:
                # Only contains and is-contained-in relationships considered
                continue
            key = (from_region.region_set_id, to_region.region_set_id)
            region_contained_in.setdefault(key, {})
            assert from_region.id is not None
            assert to_region.id is not None
            region_contained_in[key][str(from_region.id)] = str(to_region.id)

        return regions, region_set_regions_map, region_contained_in

    def _retrieve_organization_data(
        self,
    ) -> list[model.Organization]:
        app = self.case_service.app
        # Retrieve relevant organizations
        organizations: list[model.Organization] = app.handle(
            command.OrganizationCrudCommand(
                operation=CrudOperation.READ_ALL,
            )
        )
        return organizations


# TODO: for reference, remove when no longer needed
# def tfm_geo_resolution(
#     region_id_contained_in: dict[tuple[str, str], str],
#     values1: Iterable[str | None],
#     region_set_id2: str,
#     orig_values2: Iterable[str | None] | None,
# ) -> tuple[list[str | None], list[tuple[str | None, str | None]]]:
#     new_values2 = [
#         None if x is None else region_id_contained_in.get((x, region_set_id2))
#         for x in values1
#     ]
#     if orig_values2 is None:
#         diff_values = []
#     else:
#         # Only replace original values with a non-null value
#         new_values2 = [y if x is None else x for x, y in zip(new_values2, orig_values2)]
#         diff_values = [
#             (x, y)
#             for x, y in zip(orig_values2, new_values2)
#             if x is not None and y is not None and y != x
#         ]
#     return new_values2, diff_values


# def tfm_age_category(
#     age_categories1: dict[str : tuple[float, float]] | None,
#     values1: Iterable[str | None],
#     age_categories2: dict[str : tuple[float, float]] | None,
#     orig_values2: Iterable[str | None] | None,
# ) -> tuple[list[str | None], list[tuple[str | None, str | None]]]:
#     n = len(values1)
#     if not age_categories1:
#         # Variable 1 numeric
#         if not age_categories2:
#             # Variable 2 numeric -> no action, keep 2
#             new_values2 = list(orig_values2)
#         else:
#             # Variable 2 ordinal -> map to number to age category
#             new_values2 = [None] * n
#             for i, value1 in enumerate(values1):
#                 if value1 is None:
#                     continue
#                 for age_category_id2, bounds2 in age_categories2.items():
#                     if bounds2[0] <= value1 < bounds2[1]:
#                         new_values2[i] = age_category_id2
#                         break
#     else:
#         # Variable 1 ordinal
#         if not age_categories2:
#             # Variable 2 numeric -> no action, keep 2
#             new_values2 = list(orig_values2)
#         else:
#             # Variable 2 ordinal -> map to age category 1 to age category 2
#             # but only if all age categories 1 fit within an age category 2
#             # to avoid consistently removing some values and thereby introducing bias
#             age_category_map = {}
#             is_complete_mapping = True
#             for age_category1, bounds1 in age_categories1.items():
#                 for age_category2, bounds2 in age_categories2.items():
#                     if bounds1[0] >= bounds2[0] and bounds1[1] <= bounds2[1]:
#                         # Age category 1 maps to age category 2
#                         age_category_map[age_category1] = age_category2
#                         break
#                 if age_category1 not in age_category_map:
#                     is_complete_mapping = False
#             if not is_complete_mapping:
#                 new_values2 = list(orig_values2)
#             else:
#                 new_values2 = [None] * n
#                 for i, value1 in enumerate(values1):
#                     if value1 is None:
#                         continue
#                     new_values2[i] = age_category_map[value1]
#     if orig_values2 is None:
#         diff_values = []
#     else:
#         # Only replace original values with a non-null value
#         new_values2 = [y if x is None else x for x, y in zip(new_values2, orig_values2)]
#         diff_values = [
#             (x, y)
#             for x, y in zip(orig_values2, new_values2)
#             if x is not None and y is not None and y != x
#         ]
#     return new_values2, diff_values



from gen_epix.commondb.domain.repository.system import BaseSystemRepository
from gen_epix.fastapp.repositories import DictRepository


class SystemDictRepository(DictRepository, BaseSystemRepository):
    pass



import re
from typing import Any, Literal, Self

from pydantic import Field, model_validator

from gen_epix.filter.base import Filter
from gen_epix.filter.enum import FilterType


class RegexFilter(Filter):
    pattern: str = Field(description="The regular expression to match.", frozen=True)

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        try:
            self._pattern = re.compile(self.pattern)
        except re.error as exc:
            raise AssertionError("Invalid regular expression.") from exc
        return self

    def _match(self, value: Any) -> bool:
        return self._pattern.match(value) is not None


class TypedRegexFilter(RegexFilter):
    type: Literal[FilterType.REGEX.value]



"""
Object transformer implementation.
"""

from collections.abc import Callable
from typing import Any

from gen_epix.transform.adapter import ObjectAdapter
from gen_epix.transform.transformer import Transformer


class ObjectTransformer(Transformer):
    """Transform entire object using a custom function."""

    def __init__(self, transform_fn: Callable[[Any], Any], name: str | None = None):
        super().__init__(name)
        self.transform_fn = transform_fn

    def transform(self, obj: ObjectAdapter) -> ObjectAdapter:
        """Transform the entire object."""
        original = obj.unwrap()
        transformed = self.transform_fn(original)
        return ObjectAdapter(transformed)



# pylint: disable=useless-import-alias
from gen_epix import fastapp
from gen_epix.commondb.domain import enum
from gen_epix.commondb.domain.command.abac import (
    OrganizationAdminPolicyCrudCommand as OrganizationAdminPolicyCrudCommand,
)
from gen_epix.commondb.domain.command.abac import (
    RetrieveOrganizationsUnderAdminCommand as RetrieveOrganizationsUnderAdminCommand,
)
from gen_epix.commondb.domain.command.base import Command as Command
from gen_epix.commondb.domain.command.base import CrudCommand as CrudCommand
from gen_epix.commondb.domain.command.base import (
    UpdateAssociationCommand as UpdateAssociationCommand,
)
from gen_epix.commondb.domain.command.organization import (
    ContactCrudCommand as ContactCrudCommand,
)
from gen_epix.commondb.domain.command.organization import (
    DataCollectionCrudCommand as DataCollectionCrudCommand,
)
from gen_epix.commondb.domain.command.organization import (
    DataCollectionSetCrudCommand as DataCollectionSetCrudCommand,
)
from gen_epix.commondb.domain.command.organization import (
    DataCollectionSetDataCollectionUpdateAssociationCommand as DataCollectionSetDataCollectionUpdateAssociationCommand,
)
from gen_epix.commondb.domain.command.organization import (
    DataCollectionSetMemberCrudCommand as DataCollectionSetMemberCrudCommand,
)
from gen_epix.commondb.domain.command.organization import (
    IdentifierIssuerCrudCommand as IdentifierIssuerCrudCommand,
)
from gen_epix.commondb.domain.command.organization import (
    InviteUserCommand as InviteUserCommand,
)
from gen_epix.commondb.domain.command.organization import (
    OrganizationCrudCommand as OrganizationCrudCommand,
)
from gen_epix.commondb.domain.command.organization import (
    OrganizationSetCrudCommand as OrganizationSetCrudCommand,
)
from gen_epix.commondb.domain.command.organization import (
    OrganizationSetMemberCrudCommand as OrganizationSetMemberCrudCommand,
)
from gen_epix.commondb.domain.command.organization import (
    OrganizationSetOrganizationUpdateAssociationCommand as OrganizationSetOrganizationUpdateAssociationCommand,
)
from gen_epix.commondb.domain.command.organization import (
    RegisterInvitedUserCommand as RegisterInvitedUserCommand,
)
from gen_epix.commondb.domain.command.organization import (
    RetrieveInviteUserConstraintsCommand as RetrieveInviteUserConstraintsCommand,
)
from gen_epix.commondb.domain.command.organization import (
    RetrieveOrganizationAdminNameEmailsCommand as RetrieveOrganizationAdminNameEmailsCommand,
)
from gen_epix.commondb.domain.command.organization import (
    RetrieveOrganizationContactCommand as RetrieveOrganizationContactCommand,
)
from gen_epix.commondb.domain.command.organization import (
    SiteCrudCommand as SiteCrudCommand,
)
from gen_epix.commondb.domain.command.organization import (
    UpdateUserCommand as UpdateUserCommand,
)
from gen_epix.commondb.domain.command.organization import (
    UpdateUserOwnOrganizationCommand as UpdateUserOwnOrganizationCommand,
)
from gen_epix.commondb.domain.command.organization import (
    UserCrudCommand as UserCrudCommand,
)
from gen_epix.commondb.domain.command.organization import (
    UserInvitationCrudCommand as UserInvitationCrudCommand,
)
from gen_epix.commondb.domain.command.rbac import (
    RetrieveOwnPermissionsCommand as RetrieveOwnPermissionsCommand,
)
from gen_epix.commondb.domain.command.rbac import (
    RetrieveSubRolesCommand as RetrieveSubRolesCommand,
)
from gen_epix.commondb.domain.command.system import (
    OutageCrudCommand as OutageCrudCommand,
)
from gen_epix.commondb.domain.command.system import (
    RetrieveLicensesCommand as RetrieveLicensesCommand,
)
from gen_epix.commondb.domain.command.system import (
    RetrieveOutagesCommand as RetrieveOutagesCommand,
)
from gen_epix.fastapp.services.auth import (
    GetIdentityProvidersCommand as GetIdentityProvidersCommand,
)

COMMANDS_BY_SERVICE_TYPE: dict[enum.ServiceType, frozenset[type[fastapp.Command]]] = {
    enum.ServiceType.AUTH: frozenset(
        {
            GetIdentityProvidersCommand,
        }
    ),
    enum.ServiceType.ORGANIZATION: frozenset(
        {
            ContactCrudCommand,
            DataCollectionCrudCommand,
            DataCollectionSetCrudCommand,
            DataCollectionSetDataCollectionUpdateAssociationCommand,
            DataCollectionSetMemberCrudCommand,
            IdentifierIssuerCrudCommand,
            InviteUserCommand,
            OrganizationCrudCommand,
            OrganizationSetCrudCommand,
            OrganizationSetMemberCrudCommand,
            OrganizationSetOrganizationUpdateAssociationCommand,
            RegisterInvitedUserCommand,
            RetrieveInviteUserConstraintsCommand,
            RetrieveOrganizationAdminNameEmailsCommand,
            RetrieveOrganizationContactCommand,
            SiteCrudCommand,
            UpdateUserCommand,
            UpdateUserOwnOrganizationCommand,
            UserCrudCommand,
            UserInvitationCrudCommand,
        }
    ),
    enum.ServiceType.RBAC: frozenset(
        {
            RetrieveOwnPermissionsCommand,
            RetrieveSubRolesCommand,
        }
    ),
    enum.ServiceType.SYSTEM: frozenset(
        {
            OutageCrudCommand,
            RetrieveOutagesCommand,
            RetrieveLicensesCommand,
        }
    ),
    enum.ServiceType.ABAC: frozenset(
        {
            OrganizationAdminPolicyCrudCommand,
            RetrieveOrganizationsUnderAdminCommand,
        }
    ),
}



from collections.abc import Hashable
from enum import Enum

from gen_epix.fastapp import exc
from gen_epix.fastapp.domain.entity import Entity
from gen_epix.fastapp.domain.link import Link
from gen_epix.fastapp.enum import CrudOperation, PermissionType, PermissionTypeSet
from gen_epix.fastapp.model import Command, CrudCommand, Model, Permission


class Domain:
    CRUD_PERMISSION_TYPE_MAP: dict[CrudOperation, PermissionType] = {
        CrudOperation.CREATE_ONE: PermissionType.CREATE,
        CrudOperation.CREATE_SOME: PermissionType.CREATE,
        CrudOperation.READ_ALL: PermissionType.READ,
        CrudOperation.READ_SOME: PermissionType.READ,
        CrudOperation.READ_ONE: PermissionType.READ,
        CrudOperation.UPDATE_ONE: PermissionType.UPDATE,
        CrudOperation.UPDATE_SOME: PermissionType.UPDATE,
        CrudOperation.DELETE_ONE: PermissionType.DELETE,
        CrudOperation.DELETE_SOME: PermissionType.DELETE,
        CrudOperation.EXISTS_ONE: PermissionType.READ,
        CrudOperation.EXISTS_SOME: PermissionType.READ,
    }
    PERMISSION_TYPE_CRUD_MAP: dict[PermissionType, frozenset[CrudOperation]] = {
        PermissionType.CREATE: frozenset(
            {CrudOperation.CREATE_ONE, CrudOperation.CREATE_SOME}
        ),
        PermissionType.READ: frozenset(
            {
                CrudOperation.READ_ALL,
                CrudOperation.READ_SOME,
                CrudOperation.READ_ONE,
                CrudOperation.EXISTS_ONE,
                CrudOperation.EXISTS_SOME,
            }
        ),
        PermissionType.UPDATE: frozenset(
            {CrudOperation.UPDATE_ONE, CrudOperation.UPDATE_SOME}
        ),
        PermissionType.DELETE: frozenset(
            {
                CrudOperation.DELETE_ALL,
                CrudOperation.DELETE_SOME,
                CrudOperation.DELETE_ONE,
            }
        ),
    }

    @staticmethod
    def get_service_name(service_type: Hashable | None) -> str:
        if service_type is None:
            raise exc.DomainException("Service type is not given")
        if isinstance(service_type, str):
            return service_type
        if isinstance(service_type, Enum):
            return str(service_type.value)
        return str(service_type)

    @staticmethod
    def get_command_name(command_class: type[Command]) -> str:
        if command_class.NAME is None:
            command_class.NAME = command_class.__name__
        return command_class.NAME

    @staticmethod
    def get_command_permissions(command_class: type[Command]) -> frozenset[Permission]:
        if command_class._PERMISSIONS is None:
            if command_class.NAME is None:
                raise exc.InitializationServiceError(
                    f"Command {command_class} has no NAME set"
                )
            command_class._PERMISSIONS = frozenset(
                {
                    Permission(command_name=command_class.NAME, permission_type=x)
                    for x in command_class.PERMISSION_TYPE_SET.value
                }
            )
        return command_class._PERMISSIONS

    @staticmethod
    def get_model_name(model_class: type[Model]) -> str:
        if model_class.NAME is None:
            model_class.NAME = model_class.__name__
        return model_class.NAME

    def __init__(self, name: str, description: str | None = None):
        self._name = name
        self._description = description

        # Initialize sets
        self._service_types: set[Hashable] = set()
        self._entities: set[Entity] = set()
        self._models: set[type[Model]] = set()
        self._crud_commands: set[type[CrudCommand]] = set()
        self._commands: set[type[Command]] = set()
        self._permissions: set[Permission] = set()
        self._entity_dag: dict[Entity, list[Entity]] = {}
        self._dag_sorted_entities: list[Entity] = []

        # Initialize mappings between classes and their names
        self._service_type_for_name: dict[str, Hashable] = {}
        self._name_for_service_type: dict[Hashable, str] = {}
        self._command_for_name: dict[str, type[Command]] = {}
        self._name_for_command: dict[type[Command], str] = {}
        self._model_for_name: dict[str, type[Model]] = {}
        self._name_for_model: dict[type[Model], str] = {}
        self._permission_for_tuple: dict[
            tuple[str | type[Command], PermissionType], Permission
        ] = {}

        # Initialize mappings between types of classes (service_type, Entity, Model,
        # Command, Permission)
        # These are pre-computed upon registration for efficiency
        # service_type and Entity
        self._service_type_for_entity: dict[Entity, Hashable] = {}
        self._entities_for_service_type: dict[Hashable, set[Entity]] = {}
        # service type and Model
        self._service_type_for_model: dict[type[Model] | str, Hashable] = {}
        self._models_for_service_type: dict[Hashable, set[type[Model]]] = {}
        # service type and Command
        self._service_type_for_command: dict[type[Command] | str, Hashable] = {}
        self._commands_for_service_type: dict[Hashable, set[type[Command]]] = {}
        self._crud_commands_for_service_type: dict[Hashable, set[type[CrudCommand]]] = (
            {}
        )
        # service type and Permission
        self._service_type_for_permission: dict[Permission, Hashable] = {}
        self._permissions_for_service_type: dict[Hashable, set[Permission]] = {}
        # Entity and Model
        self._entity_for_model: dict[type[Model] | str, Entity] = {}
        self._model_for_entity: dict[Entity, type[Model]] = {}
        # Entity and CrudCommand
        self._entity_for_crud_command: dict[type[CrudCommand] | str, Entity] = {}
        self._crud_command_for_entity: dict[Entity, type[CrudCommand]] = {}
        # Entity and Permission
        self._entity_for_permission: dict[Permission, Entity] = {}
        self._permissions_for_entity: dict[Entity, frozenset[Permission]] = {}
        # Model and CrudCommand
        self._crud_command_for_model: dict[type[Model] | str, type[CrudCommand]] = {}
        self._model_for_crud_command: dict[type[CrudCommand] | str, type[Model]] = {}
        # Model and Permission
        self._model_for_permission: dict[Permission, type[Model]] = {}
        self._permissions_for_model: dict[type[Model] | str, frozenset[Permission]] = {}
        # Command and Permission
        self._permissions_for_command: dict[
            type[Command] | str, frozenset[Permission]
        ] = {}
        self._command_for_permission: dict[Permission, type[Command]] = {}

    @property
    def name(self) -> str:
        return self._name

    @property
    def description(self) -> str | None:
        return self._description

    @property
    def service_types(self) -> frozenset[Hashable]:
        return frozenset(self._service_types)

    @property
    def service_names(self) -> frozenset[str]:
        return frozenset(self._service_type_for_name.keys())

    @property
    def entities(self) -> frozenset[Entity]:
        return frozenset(self._entities)

    @property
    def commands(self) -> frozenset[type[Command]]:
        return frozenset(self._commands)

    @property
    def command_names(self) -> frozenset[str]:
        return frozenset(self._command_for_name.keys())

    @property
    def crud_commands(self) -> frozenset[type[CrudCommand]]:
        return frozenset(self._crud_commands)

    @property
    def models(self) -> frozenset[type[Model]]:
        return frozenset(self._models)

    @property
    def permissions(self) -> frozenset[Permission]:
        return frozenset(self._permissions)

    def get_commands(
        self, include_crud: bool = False, frozen: bool = True
    ) -> set[type[Command]] | frozenset[type[Command]]:
        if include_crud:
            if frozen:
                return frozenset(self._commands)
            return set(self._commands)
        commands = {x for x in self._commands if x not in self._crud_commands}
        if frozen:
            return frozenset(commands)
        return set(commands)

    def get_service_type_for_entity(
        self, entity: Entity, verify: bool = False
    ) -> Hashable | None:
        if verify:
            self._verify_entity_exists(entity)
            return self._service_type_for_entity[entity]
        return self._service_type_for_entity.get(entity)

    def get_entities_for_service_type(
        self, service_type: Hashable, verify: bool = False, frozen: bool = True
    ) -> set[Entity] | frozenset[Entity] | None:
        if verify:
            self._verify_service_type_exists(service_type)
        entities = self._entities_for_service_type.get(service_type)
        if entities is None:
            return None
        return frozenset(entities) if frozen else entities

    def get_service_types(self) -> set[Hashable]:
        return set(self._service_types)

    def get_service_type_for_model(
        self, model_class: type[Model], verify: bool = False
    ) -> Hashable:
        if verify:
            self._verify_model_exists(model_class)
        return self._service_type_for_model.get(model_class)

    def get_models_for_service_type(
        self, service_type: Hashable, frozen: bool = True
    ) -> set[type[Model]] | frozenset[type[Model]]:
        self._verify_service_type_exists(service_type)
        if frozen:
            return frozenset(self._models_for_service_type[service_type])
        return set(self._models_for_service_type[service_type])

    def get_service_type_for_command(self, command_class: type[Command]) -> Hashable:
        self._verify_command_exists(command_class)
        return self._service_type_for_command[command_class]

    def get_commands_for_service_type(
        self,
        service_type: Hashable,
        frozen: bool = True,
        base_class: type[Command] | None = None,
    ) -> set[type[Command]] | frozenset[type[Command]]:
        self._verify_service_type_exists(service_type)
        if frozen:
            if base_class:
                return frozenset(
                    {
                        x
                        for x in self._commands_for_service_type[service_type]
                        if issubclass(x, base_class)
                    }
                )
            return frozenset(self._commands_for_service_type[service_type])
        if base_class:
            return {
                x
                for x in self._commands_for_service_type[service_type]
                if issubclass(x, base_class)
            }
        return set(self._commands_for_service_type[service_type])

    def get_crud_commands_for_service_type(
        self,
        service_type: Hashable,
        frozen: bool = True,
        base_class: type[CrudCommand] | None = None,
    ) -> set[type[CrudCommand]] | frozenset[type[CrudCommand]]:
        self._verify_service_type_exists(service_type)
        if frozen:
            if base_class:
                return frozenset(
                    {
                        x
                        for x in self._crud_commands_for_service_type[service_type]
                        if issubclass(x, base_class)
                    }
                )
            return frozenset(self._crud_commands_for_service_type[service_type])
        if base_class:
            return frozenset(
                {
                    x
                    for x in self._crud_commands_for_service_type[service_type]
                    if issubclass(x, base_class)
                }
            )
        return frozenset(self._crud_commands_for_service_type[service_type])

    def get_command_for_name(self, command_name: str) -> type[Command]:
        self._verify_command_exists(command_name)
        return self._command_for_name[command_name]

    def get_permissions_for_service_type(
        self, service_type: Hashable, frozen: bool = True
    ) -> set[Permission] | frozenset[Permission]:
        self._verify_service_type_exists(service_type)
        if frozen:
            return frozenset(self._permissions_for_service_type[service_type])
        return set(self._permissions_for_service_type[service_type])

    def get_permissions_for_domain(
        self, frozen: bool = True
    ) -> set[Permission] | frozenset[Permission]:
        """
        Get permissions for all the commands in the domain.
        """
        if frozen:
            return frozenset(self._permissions)
        return set(self._permissions)

    def get_model_excluded_permissions(self) -> dict[type[Model], PermissionTypeSet]:
        """
        For all registered CRUD commands, return a dict where the key is the
        model class and the value is a PermissionTypeSet that indicates which
        CRUD operations are not allowed. Only models that have missing
        permissions are included in the dict.
        """
        permission_type_map: dict[frozenset[PermissionType], PermissionTypeSet] = {
            x.value: x for x in PermissionTypeSet
        }
        excluded_permissions = {}
        for crud_command_class in self._crud_commands:
            assert crud_command_class._PERMISSIONS is not None
            curr_excluded_permissions = PermissionTypeSet.CRUD.value - {
                x.permission_type for x in crud_command_class._PERMISSIONS
            }
            if not curr_excluded_permissions:
                continue
            excluded_permissions[self._model_for_crud_command[crud_command_class]] = (
                permission_type_map[frozenset(curr_excluded_permissions)]
            )
        return excluded_permissions

    def get_service_type_for_permission(self, permission: Permission) -> Hashable:
        self._verify_permission_exists(permission)
        return self._service_type_for_permission[permission]

    def get_entity_for_model(self, model_class: type[Model]) -> Entity:
        self._verify_model_exists(model_class)
        return self._entity_for_model[model_class]

    def get_model_for_entity(self, entity: Entity) -> type[Model]:
        self._verify_entity_exists(entity)
        return self._model_for_entity[entity]

    def get_entity_for_crud_command(
        self, crud_command_class: type[CrudCommand]
    ) -> Entity:
        self._verify_command_exists(crud_command_class)
        return self._entity_for_crud_command[crud_command_class]

    def get_crud_command_for_entity(self, entity: Entity) -> type[CrudCommand]:
        self._verify_entity_exists(entity)
        return self._crud_command_for_entity[entity]

    def get_entity_for_permission(self, permission: Permission) -> Entity:
        self._verify_permission_exists(permission)
        return self._entity_for_permission[permission]

    def get_permissions_for_entity(
        self, entity: Entity, frozen: bool = True
    ) -> set[Permission] | frozenset[Permission]:
        self._verify_entity_exists(entity)
        if frozen:
            return frozenset(self._permissions_for_entity[entity])
        return set(self._permissions_for_entity[entity])

    def get_crud_command_for_model(self, model_class: type[Model]) -> type[CrudCommand]:
        self._verify_model_exists(model_class)
        return self._crud_command_for_model[model_class]

    def get_model_for_crud_command(
        self, crud_command_class: type[CrudCommand]
    ) -> type[Model]:
        self._verify_command_exists(crud_command_class)
        return self._model_for_crud_command[crud_command_class]

    def get_model_for_permission(self, permission: Permission) -> type[Model]:
        self._verify_permission_exists(permission)
        return self._model_for_permission[permission]

    def get_permissions_for_model(
        self,
        model_class: type[Model],
        frozen: bool = True,
        permission_type_set: PermissionTypeSet | None = None,
    ) -> set[Permission] | frozenset[Permission]:
        self._verify_model_exists(model_class)
        return self.get_permissions_for_command(
            self.get_crud_command_for_model(model_class),
            frozen=frozen,
            permission_type_set=permission_type_set,
        )

    def get_permissions_for_command(
        self,
        command_class: type[Command],
        frozen: bool = True,
        permission_type_set: PermissionTypeSet | None = None,
    ) -> set[Permission] | frozenset[Permission]:
        self._verify_command_exists(command_class)
        if permission_type_set:
            permissions = set(
                {
                    x
                    for x in self._permissions_for_command[command_class]
                    if x.permission_type in permission_type_set.value
                }
            )
            if frozen:
                return frozenset(permissions)
            return permissions
        if frozen:
            return self._permissions_for_command[command_class]
        return set(self._permissions_for_command[command_class])

    def get_command_for_permission(self, permission: Permission) -> type[Command]:
        self._verify_permission_exists(permission)
        return self._command_for_permission[permission]

    def get_permission(
        self,
        command_class_or_name: type[Command] | str,
        permission_type: PermissionType,
    ) -> Permission:
        self._verify_command_exists(command_class_or_name)
        return self._permission_for_tuple[(command_class_or_name, permission_type)]

    def get_permission_for_command_instance(
        self,
        cmd: Command,
    ) -> Permission:
        command_class = type(cmd)
        if issubclass(command_class, CrudCommand):
            permission_type = Domain.CRUD_PERMISSION_TYPE_MAP[cmd.operation]  # type: ignore
        else:
            permission_type = PermissionType.EXECUTE
        return self._permission_for_tuple[(command_class, permission_type)]

    def get_model_links(
        self,
        model_class: type[Model],
        as_tuple: bool = False,
        service_type: Hashable | None = None,
        url_name: str | None = None,
        database_name: str | None = None,
        schema_name: str | None = None,
        invert: bool = False,
    ) -> dict[int, tuple[str, type[Model], str | None]] | dict[int, Link]:
        self._verify_model_exists(model_class)
        entity = self.get_entity_for_model(model_class)
        links = {}
        for link_type_id, link in entity.links.items():
            link_entity = self.get_entity_for_model(link.link_model_class)  # type: ignore
            if (
                service_type
                and (self.get_service_type_for_entity(link_entity) != service_type)
                != invert
            ):
                continue
            if url_name and (link_entity.url_name != url_name) != invert:
                continue
            if database_name and (link_entity.database_name != database_name) != invert:
                continue
            if schema_name and (link_entity.schema_name != schema_name) != invert:
                continue
            links[link_type_id] = link
        if as_tuple:
            return {x: y.to_tuple() for x, y in links.items()}  # type: ignore
        return links

    def get_dag_sorted_entities(
        self,
        service_type: Hashable | None = None,
        persistable: bool | None = None,
        url_name: str | None = None,
        database_name: str | None = None,
        schema_name: str | None = None,
        invert: bool = False,
        reverse: bool = False,
    ) -> list[Entity]:
        if service_type:
            self._verify_service_type_exists(service_type)
        entities = []
        for entity in self._dag_sorted_entities:
            if (
                service_type
                and (self.get_service_type_for_entity(entity) != service_type) != invert
            ):
                continue
            if (
                persistable is not None
                and (entity.persistable != persistable) != invert
            ):
                continue
            if url_name and (entity.url_name != url_name) != invert:
                continue
            if database_name and (entity.database_name != database_name) != invert:
                continue
            if schema_name and (entity.schema_name != schema_name) != invert:
                continue
            entities.append(entity)
        if reverse:
            return list(reversed(entities))
        return entities

    def get_dag_sorted_models(
        self,
        service_type: Hashable | None = None,
        persistable: bool | None = None,
        url_name: str | None = None,
        database_name: str | None = None,
        schema_name: str | None = None,
        invert: bool = False,
        reverse: bool = False,
    ) -> list[type[Model]]:
        return [
            x.model_class  # type: ignore
            for x in self.get_dag_sorted_entities(
                service_type=service_type,
                persistable=persistable,
                url_name=url_name,
                database_name=database_name,
                schema_name=schema_name,
                invert=invert,
                reverse=reverse,
            )
        ]

    def register_service_type(self, service_type: Hashable) -> Hashable:
        if service_type not in self._service_types:
            self._service_types.add(service_type)
            self._entities_for_service_type[service_type] = set()
            self._commands_for_service_type[service_type] = set()
            self._crud_commands_for_service_type[service_type] = set()
            self._models_for_service_type[service_type] = set()
            self._permissions_for_service_type[service_type] = set()
        return service_type

    def register_entity(
        self,
        entity: Entity,
        service_type: Hashable | None = None,
        model_class: type[Model] | None = None,
        crud_command_class: type[CrudCommand] | None = None,
        on_cycle: str = "raise",
    ) -> Entity:
        # Set service type Model and CrudCommand
        if model_class:
            entity.set_model_class(model_class)
        if crud_command_class:
            entity.set_crud_command_class(crud_command_class)

        # Register new service_type
        self.register_service_type(service_type)

        # Register new Entity
        if entity not in self._entities:
            model_class = entity.model_class  # type: ignore
            if model_class is None:
                raise exc.InitializationServiceError(
                    f"Model class not set for entity {entity.id}"
                )
            model_name = Domain.get_model_name(model_class)
            if model_name in self._model_for_name:
                raise exc.DomainException(
                    f"Model name {model_name} is already registered"
                )
            # Add Entity and Model
            self._entities.add(entity)
            self._entity_dag[entity] = []
            self._models.add(model_class)
            self._model_for_name[model_name] = model_class
            self._name_for_model[model_class] = model_name

            # Add relations between classes (service_type, Entity, Model)
            self._service_type_for_entity[entity] = service_type
            self._entities_for_service_type[service_type].add(entity)
            self._service_type_for_model[model_class] = service_type
            self._service_type_for_model[model_name] = service_type
            self._models_for_service_type[service_type].add(model_class)
            self._entity_for_model[model_class] = entity
            self._entity_for_model[model_name] = entity
            self._model_for_entity[entity] = model_class

            # Update entity DAG
            for link in entity.links.values():
                if link.link_model_class not in self._models:
                    if on_cycle.upper() == "RAISE":
                        raise exc.DomainException(
                            f"Entity {entity.name} references unknown model {link.link_model_class} - add entities in DAG sorted order"
                        )
                    if on_cycle.upper() != "IGNORE":
                        continue
                    raise NotImplementedError(f"Unsupported on_cycle: {on_cycle}")
                link_entity = self.get_entity_for_model(link.link_model_class)  # type: ignore
                self._entity_dag[entity].append(link_entity)
            self._dag_sorted_entities.append(entity)

        # Register CrudCommand if necessary
        if entity.persistable and entity.crud_command_class:
            self.register_command(entity.crud_command_class, service_type=service_type)  # type: ignore

        return entity

    def register_command(
        self, command_class: type[Command], service_type: Hashable | None = None
    ) -> type[Command]:
        command_name = Domain.get_command_name(command_class)

        # Verify already registered Command
        if command_class in self._commands:
            if command_name != self._name_for_command[command_class]:
                raise exc.DomainException(
                    f"Command {command_name} is already registered with different name {self._name_for_command[command_class]}"
                )
            linked_service_type = self._service_type_for_command[command_class]
            if linked_service_type != service_type:
                raise exc.DomainException(
                    f"Command {command_name} is already registered with service type {linked_service_type}"
                )
            if issubclass(command_class, CrudCommand):
                if command_class.MODEL_CLASS is None:
                    raise exc.DomainException(
                        f"Model class not set for CrudCommand {command_name}"
                    )
                linked_model_class = self._model_for_crud_command[command_class]
                if linked_model_class is not command_class.MODEL_CLASS:
                    raise exc.DomainException(
                        f"Command {command_name} is already registered and linked to model {linked_model_class.NAME}"
                    )
            return command_class

        # Register new service_type
        self.register_service_type(service_type)

        # Register new Command
        permissions = Domain.get_command_permissions(command_class)
        self._commands.add(command_class)
        self._permissions.update(permissions)
        self._command_for_name[command_name] = command_class
        self._name_for_command[command_class] = command_name
        # Add relations between classes (service_type, Command, Permission)
        self._service_type_for_command[command_class] = service_type
        self._commands_for_service_type[service_type].add(command_class)
        self._permissions_for_service_type[service_type].update(permissions)
        self._permissions_for_command[command_class] = permissions
        self._permissions_for_command[command_name] = permissions
        for permission in permissions:
            self._service_type_for_permission[permission] = service_type
            self._command_for_permission[permission] = command_class
            self._permission_for_tuple[(command_class, permission.permission_type)] = (
                permission
            )
            self._permission_for_tuple[(command_name, permission.permission_type)] = (
                permission
            )

        # Add CrudCommand specific verifications and data
        if issubclass(command_class, CrudCommand):
            crud_command_class = command_class
            # Verify model class
            model_class = crud_command_class.MODEL_CLASS
            if not model_class:
                raise exc.DomainException(
                    f"Model class not set for CRUD command {command_name}"
                )
            # Verify entity
            entity = model_class.ENTITY
            if not entity:
                raise exc.DomainException(
                    f"Entity not set for model {Domain.get_model_name(model_class)}"
                )
            if not entity.persistable:
                raise exc.DomainException(f"Entity {entity.name} is not persistable")
            # Set entity Model and CrudCommand if necessary
            if not entity.has_model():
                entity.set_model_class(model_class)
            if not entity.crud_command_class:
                entity.set_crud_command_class(crud_command_class)

            # Register new CrudCommand
            self._crud_commands.add(crud_command_class)
            # Add relations between classes (service_type, Entity, Model, CrudCommand, Permission)
            model_name = Domain.get_model_name(model_class)
            self._crud_commands_for_service_type[service_type].add(crud_command_class)
            self._entity_for_crud_command[crud_command_class] = entity
            self._entity_for_crud_command[command_name] = entity
            self._crud_command_for_entity[entity] = crud_command_class
            self._model_for_crud_command[crud_command_class] = model_class
            self._model_for_crud_command[command_name] = model_class
            self._permissions_for_model[model_class] = permissions
            for permission in permissions:
                self._model_for_permission[permission] = model_class
            self._crud_command_for_model[model_class] = crud_command_class
            self._crud_command_for_model[model_name] = crud_command_class

            # Set entity to be registered
            if entity not in self._entities:
                self.register_entity(
                    entity,
                    model_class=model_class,
                    crud_command_class=crud_command_class,
                )
        return command_class

    def _verify_model_has_entity(self, model_class: type[Model]) -> None:
        if model_class.ENTITY is None:
            raise exc.DomainException(
                f"Entity not set for model {Domain.get_model_name(model_class)}"
            )

    def _verify_entity_exists(self, entity: Entity) -> None:
        if entity not in self._entities:
            raise exc.DomainException(f"Entity {entity.name} is not registered")

    def _verify_service_type_exists(self, service_type: Hashable) -> None:
        if service_type not in self._service_types:
            raise exc.DomainException(f"Service type {service_type} is not registered")

    def _verify_command_exists(
        self, command_class_or_name: type[Command] | str
    ) -> None:
        if isinstance(command_class_or_name, str):
            command_name = command_class_or_name
            if command_name not in self._command_for_name:
                raise exc.DomainException(f"Command {command_name} is not registered")
        else:
            command_class = command_class_or_name
            if command_class not in self._commands:
                command_name = Domain.get_command_name(command_class)
                raise exc.DomainException(f"Command {command_name} is not registered")

    def _verify_model_exists(self, model_class_or_name: type[Model] | str) -> None:
        if isinstance(model_class_or_name, str):
            model_name = model_class_or_name
            if model_name not in self._model_for_name:
                raise exc.DomainException(f"Model {model_name} is not registered")
        else:
            model_class = model_class_or_name
            if model_class not in self._models:
                raise exc.DomainException(
                    f"Model {Domain.get_model_name(model_class)} is not registered"
                )

    def _verify_permission_exists(self, permission: Permission) -> None:
        if permission not in self._permissions:
            raise exc.DomainException(f"Permission {permission} is not registered")



from uuid import UUID

from pydantic import Field

from gen_epix import fastapp


class Model(fastapp.Model):
    id: UUID | None = Field(
        default=None,
        description="The unique identifier for the obj.",
    )



from decimal import Decimal
from typing import Annotated, Literal

from pydantic import Field, WithJsonSchema

from gen_epix.filter.enum import FilterType
from gen_epix.filter.range import RangeFilter


class NumberRangeFilter(RangeFilter):
    lower_bound: (
        Annotated[
            int | float | Decimal,
            WithJsonSchema({"type": "number"}),
        ]
        | None
    ) = Field(default=None, description="The lower bound of the range.", frozen=True)
    upper_bound: (
        Annotated[
            int | float | Decimal,
            WithJsonSchema({"type": "number"}),
        ]
        | None
    ) = Field(default=None, description="The upper bound of the range.", frozen=True)


class TypedNumberRangeFilter(NumberRangeFilter):
    type: Literal[FilterType.NUMBER_RANGE.value]



# pylint: disable=useless-import-alias
from gen_epix.fastapp.repositories.dict import DictRepository as DictRepository
from gen_epix.fastapp.repositories.dict import DictUnitOfWork as DictUnitOfWork
from gen_epix.fastapp.repositories.sa import SAMapper as SAMapper
from gen_epix.fastapp.repositories.sa import SARepository as SARepository
from gen_epix.fastapp.repositories.sa import SAUnitOfWork as SAUnitOfWork
from gen_epix.fastapp.repositories.sa import (
    ServerUtcCurrentTime as ServerUtcCurrentTime,
)
from gen_epix.fastapp.repositories.sa import ServerUtcTimestamp as ServerUtcTimestamp
from gen_epix.fastapp.repositories.sa import (
    create_sa_type_from_field_info as create_sa_type_from_field_info,
)



# pylint: disable=useless-import-alias
from gen_epix.commondb.domain import DOMAIN, enum, model
from gen_epix.commondb.repositories.sa_model.abac import (
    OrganizationAdminPolicy as OrganizationAdminPolicy,
)
from gen_epix.commondb.repositories.sa_model.abac import (
    OrganizationAdminPolicyMixin as OrganizationAdminPolicyMixin,
)
from gen_epix.commondb.repositories.sa_model.base import (
    DB_METADATA_FIELDS as DB_METADATA_FIELDS,
)
from gen_epix.commondb.repositories.sa_model.base import (
    GENERATE_SERVICE_METADATA as GENERATE_SERVICE_METADATA,
)
from gen_epix.commondb.repositories.sa_model.base import (
    SERVICE_METADATA_FIELDS as SERVICE_METADATA_FIELDS,
)
from gen_epix.commondb.repositories.sa_model.base import (
    NoIdRowMetadataMixin as NoIdRowMetadataMixin,
)
from gen_epix.commondb.repositories.sa_model.base import (
    RowMetadataMixin as RowMetadataMixin,
)
from gen_epix.commondb.repositories.sa_model.organization import Contact as Contact
from gen_epix.commondb.repositories.sa_model.organization import (
    ContactMixin as ContactMixin,
)
from gen_epix.commondb.repositories.sa_model.organization import (
    DataCollection as DataCollection,
)
from gen_epix.commondb.repositories.sa_model.organization import (
    DataCollectionMixin as DataCollectionMixin,
)
from gen_epix.commondb.repositories.sa_model.organization import (
    DataCollectionSet as DataCollectionSet,
)
from gen_epix.commondb.repositories.sa_model.organization import (
    DataCollectionSetMember as DataCollectionSetMember,
)
from gen_epix.commondb.repositories.sa_model.organization import (
    DataCollectionSetMemberMixin as DataCollectionSetMemberMixin,
)
from gen_epix.commondb.repositories.sa_model.organization import (
    DataCollectionSetMixin as DataCollectionSetMixin,
)
from gen_epix.commondb.repositories.sa_model.organization import (
    IdentifierIssuer as IdentifierIssuer,
)
from gen_epix.commondb.repositories.sa_model.organization import (
    IdentifierIssuerMixin as IdentifierIssuerMixin,
)
from gen_epix.commondb.repositories.sa_model.organization import (
    Organization as Organization,
)
from gen_epix.commondb.repositories.sa_model.organization import (
    OrganizationMixin as OrganizationMixin,
)
from gen_epix.commondb.repositories.sa_model.organization import (
    OrganizationSet as OrganizationSet,
)
from gen_epix.commondb.repositories.sa_model.organization import (
    OrganizationSetMember as OrganizationSetMember,
)
from gen_epix.commondb.repositories.sa_model.organization import (
    OrganizationSetMemberMixin as OrganizationSetMemberMixin,
)
from gen_epix.commondb.repositories.sa_model.organization import (
    OrganizationSetMixin as OrganizationSetMixin,
)
from gen_epix.commondb.repositories.sa_model.organization import Site as Site
from gen_epix.commondb.repositories.sa_model.organization import SiteMixin as SiteMixin
from gen_epix.commondb.repositories.sa_model.organization import User as User
from gen_epix.commondb.repositories.sa_model.organization import (
    UserInvitation as UserInvitation,
)
from gen_epix.commondb.repositories.sa_model.organization import (
    UserInvitationMixin as UserInvitationMixin,
)
from gen_epix.commondb.repositories.sa_model.organization import UserMixin as UserMixin
from gen_epix.commondb.repositories.sa_model.system import Outage as Outage
from gen_epix.commondb.repositories.sa_model.system import OutageMixin as OutageMixin
from gen_epix.commondb.repositories.sa_model.util import create_field_metadata
from gen_epix.commondb.repositories.sa_model.util import (
    create_field_metadata as create_field_metadata,
)
from gen_epix.commondb.repositories.sa_model.util import (
    create_mapped_column as create_mapped_column,
)
from gen_epix.commondb.repositories.sa_model.util import (
    create_table_args as create_table_args,
)
from gen_epix.commondb.repositories.sa_model.util import (
    get_mixin_mapped_column as get_mixin_mapped_column,
)
from gen_epix.commondb.repositories.sa_model.util import (
    set_entity_repository_model_classes,
)
from gen_epix.commondb.repositories.sa_model.util import (
    set_entity_repository_model_classes as set_entity_repository_model_classes,
)

SA_MODELS_BY_SERVICE_TYPE: dict[enum.ServiceType, dict[type[model.Model], type]] = {
    enum.ServiceType.ABAC: {
        model.OrganizationAdminPolicy: OrganizationAdminPolicy,
    },
    enum.ServiceType.ORGANIZATION: {
        model.Contact: Contact,
        model.DataCollection: DataCollection,
        model.DataCollectionSet: DataCollectionSet,
        model.DataCollectionSetMember: DataCollectionSetMember,
        model.IdentifierIssuer: IdentifierIssuer,
        model.Organization: Organization,
        model.OrganizationSet: OrganizationSet,
        model.OrganizationSetMember: OrganizationSetMember,
        model.Site: Site,
        model.User: User,
        model.UserInvitation: UserInvitation,
    },
    enum.ServiceType.SYSTEM: {
        model.Outage: Outage,
    },
}

FIELD_NAME_MAP: dict[type, dict[str, str]] = {}

set_entity_repository_model_classes(
    DOMAIN,
    SA_MODELS_BY_SERVICE_TYPE,
    RowMetadataMixin,
    field_name_map=FIELD_NAME_MAP,
)

SERVICE_METADATA_FIELDS, DB_METADATA_FIELDS, GENERATE_SERVICE_METADATA = (
    create_field_metadata(DOMAIN)
)
